/*  Pcsx - Pc Psx Emulator
 *  Copyright (C) 1999-2002  Pcsx Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h> 
#include <stdio.h> 
#include <string.h>
#include <stdarg.h>
#include <ctype.h>

#include "Common.h"
//We try to emulate bios :) HELP US :P

char *biosA0n[256] = {
// 0x00
	"open",		"lseek",	"read",		"write",
	"close",	"ioctl",	"exit",		"sys_a0_07",
	"getc",		"putc",		"todigit",	"atof",
	"strtoul",	"strtol",	"abs",		"labs",
// 0x10
	"atoi",		"atol",		"atob",		"setjmp",
	"longjmp",	"strcat",	"strncat",	"strcmp",
	"strncmp",	"strcpy",	"strncpy",	"strlen",
	"index",	"rindex",	"strchr",	"strrchr",
// 0x20
	"strpbrk",	"strspn",	"strcspn",	"strtok",
	"strstr",	"toupper",	"tolower",	"bcopy",
	"bzero",	"bcmp",		"memcpy",	"memset",
	"memmove",	"memcmp",	"memchr",	"rand",
// 0x30
	"srand",	"qsort",	"strtod",	"malloc",
	"free",		"lsearch",	"bsearch",	"calloc",
	"realloc",	"InitHeap",	"_exit",	"getchar",
	"putchar",	"gets",		"puts",		"printf",
// 0x40
	"sys_a0_40",		"LoadTest",					"Load",		"Exec",
	"FlushCache",		"InstallInterruptHandler",	"GPU_dw",	"mem2vram",
	"SendGPUStatus",	"GPU_cw",					"GPU_cwb",	"SendPackets",
	"sys_a0_4c",		"GetGPUStatus",				"GPU_sync",	"sys_a0_4f",
// 0x50
	"sys_a0_50",		"LoadExec",				"GetSysSp",		"sys_a0_53",
	"_96_init()",		"_bu_init()",			"_96_remove()",	"sys_a0_57",
	"sys_a0_58",		"sys_a0_59",			"sys_a0_5a",	"dev_tty_init",
	"dev_tty_open",		"sys_a0_5d",			"dev_tty_ioctl","dev_cd_open",
// 0x60
	"dev_cd_read",		"dev_cd_close",			"dev_cd_firstfile",	"dev_cd_nextfile",
	"dev_cd_chdir",		"dev_card_open",		"dev_card_read",	"dev_card_write",
	"dev_card_close",	"dev_card_firstfile",	"dev_card_nextfile","dev_card_erase",
	"dev_card_undelete","dev_card_format",		"dev_card_rename",	"dev_card_6f",
// 0x70
	"_bu_init",			"_96_init",		"_96_remove",		"sys_a0_73",
	"sys_a0_74",		"sys_a0_75",	"sys_a0_76",		"sys_a0_77",
	"_96_CdSeekL",		"sys_a0_79",	"sys_a0_7a",		"sys_a0_7b",
	"_96_CdGetStatus",	"sys_a0_7d",	"_96_CdRead",		"sys_a0_7f",
// 0x80
	"sys_a0_80",		"sys_a0_81",	"sys_a0_82",		"sys_a0_83",
	"sys_a0_84",		"_96_CdStop",	"sys_a0_86",		"sys_a0_87",
	"sys_a0_88",		"sys_a0_89",	"sys_a0_8a",		"sys_a0_8b",
	"sys_a0_8c",		"sys_a0_8d",	"sys_a0_8e",		"sys_a0_8f",
// 0x90
	"sys_a0_90",		"sys_a0_91",	"sys_a0_92",		"sys_a0_93",
	"sys_a0_94",		"sys_a0_95",	"AddCDROMDevice",	"AddMemCardDevide",
	"DisableKernelIORedirection",		"EnableKernelIORedirection", "sys_a0_9a", "sys_a0_9b",
	"SetConf",			"GetConf",		"sys_a0_9e",		"SetMem",
// 0xa0
	"_boot",			"SystemError",	"EnqueueCdIntr",	"DequeueCdIntr",
	"sys_a0_a4",		"ReadSector",	"get_cd_status",	"bufs_cb_0",
	"bufs_cb_1",		"bufs_cb_2",	"bufs_cb_3",		"_card_info",
	"_card_load",		"_card_auto",	"bufs_cd_4",		"sys_a0_af",
// 0xb0
	"sys_a0_b0",		"sys_a0_b1",	"do_a_long_jmp",	"sys_a0_b3",
	"?? sub_function",
};

char *biosB0n[256] = {
// 0x00
	"SysMalloc",		"sys_b0_01",	"sys_b0_02",	"sys_b0_03",
	"sys_b0_04",		"sys_b0_05",	"sys_b0_06",	"DeliverEvent",
	"OpenEvent",		"CloseEvent",	"WaitEvent",	"TestEvent",
	"EnableEvent",		"DisableEvent",	"OpenTh",		"CloseTh",
// 0x10
	"ChangeTh",			"sys_b0_11",	"InitPAD",		"StartPAD",
	"StopPAD",			"PAD_init",		"PAD_dr",		"ReturnFromExecption",
	"ResetEntryInt",	"HookEntryInt",	"sys_b0_1a",	"sys_b0_1b",
	"sys_b0_1c",		"sys_b0_1d",	"sys_b0_1e",	"sys_b0_1f",
// 0x20
	"UnDeliverEvent",	"sys_b0_21",	"sys_b0_22",	"sys_b0_23",
	"sys_b0_24",		"sys_b0_25",	"sys_b0_26",	"sys_b0_27",
	"sys_b0_28",		"sys_b0_29",	"sys_b0_2a",	"sys_b0_2b",
	"sys_b0_2c",		"sys_b0_2d",	"sys_b0_2e",	"sys_b0_2f",
// 0x30
	"sys_b0_30",		"sys_b0_31",	"open",			"lseek",
	"read",				"write",		"close",		"ioctl",
	"exit",				"sys_b0_39",	"getc",			"putc",
	"getchar",			"putchar",		"gets",			"puts",
// 0x40
	"cd",				"format",		"firstfile",	"nextfile",
	"rename",			"delete",		"undelete",		"AddDevice",
	"RemoteDevice",		"PrintInstalledDevices", "InitCARD", "StartCARD",
	"StopCARD",			"sys_b0_4d",	"_card_write",	"_card_read",
// 0x50
	"_new_card",		"Krom2RawAdd",	"sys_b0_52",	"sys_b0_53",
	"_get_errno",		"_get_error",	"GetC0Table",	"GetB0Table",
	"_card_chan",		"sys_b0_59",	"sys_b0_5a",	"ChangeClearPAD",
	"_card_status",		"_card_wait",
};

char *biosC0n[256] = {
// 0x00
	"InitRCnt",			  "InitException",		"SysEnqIntRP",		"SysDeqIntRP",
	"get_free_EvCB_slot", "get_free_TCB_slot",	"ExceptionHandler",	"InstallExeptionHandler",
	"SysInitMemory",	  "SysInitKMem",		"ChangeClearRCnt",	"SystemError",
	"InitDefInt",		  "sys_c0_0d",			"sys_c0_0e",		"sys_c0_0f",
// 0x10
	"sys_c0_10",		  "sys_c0_11",			"InstallDevices",	"FlushStfInOutPut",
	"sys_c0_14",		  "_cdevinput",			"_cdevscan",		"_circgetc",
	"_circputc",		  "ioabort",			"sys_c0_1a",		"KernelRedirect",
	"PatchAOTable",
};

//#define r0 (cpuRegs.GPR.n.r0)
#define at (cpuRegs.GPR.n.at)
#define v0 (cpuRegs.GPR.n.v0)
#define v1 (cpuRegs.GPR.n.v1)
#define a0 (cpuRegs.GPR.n.a0)
#define a1 (cpuRegs.GPR.n.a1)
#define a2 (cpuRegs.GPR.n.a2)
#define a3 (cpuRegs.GPR.n.a3)
#define t0 (cpuRegs.GPR.n.t0)
#define t1 (cpuRegs.GPR.n.t1)
#define t2 (cpuRegs.GPR.n.t2)
#define t3 (cpuRegs.GPR.n.t3)
#define t4 (cpuRegs.GPR.n.t4)
#define t5 (cpuRegs.GPR.n.t5)
#define t6 (cpuRegs.GPR.n.t6)
#define t7 (cpuRegs.GPR.n.t7)
#define s0 (cpuRegs.GPR.n.s0)
#define s1 (cpuRegs.GPR.n.s1)
#define s2 (cpuRegs.GPR.n.s2)
#define s3 (cpuRegs.GPR.n.s3)
#define s4 (cpuRegs.GPR.n.s4)
#define s5 (cpuRegs.GPR.n.s5)
#define s6 (cpuRegs.GPR.n.s6)
#define s7 (cpuRegs.GPR.n.s7)
#define t8 (cpuRegs.GPR.n.t6)
#define t9 (cpuRegs.GPR.n.t7)
#define k0 (cpuRegs.GPR.n.k0)
#define k1 (cpuRegs.GPR.n.k1)
#define gp (cpuRegs.GPR.n.gp)
#define sp (cpuRegs.GPR.n.sp)
#define s8 (cpuRegs.GPR.n.s8)
#define ra (cpuRegs.GPR.n.ra)
#define pc0 (cpuRegs.pc)

#define Ra0 (&RAM[a0 & 0x1fffff])
#define Ra1 (&RAM[a1 & 0x1fffff])
#define Ra2 (&RAM[a2 & 0x1fffff])
#define Ra3 (&RAM[a3 & 0x1fffff])
#define Rv0 (&RAM[v0 & 0x1fffff])
#define Rsp (&RAM[sp & 0x1fffff])



typedef struct _malloc_chunk {
	unsigned long stat;
	unsigned long size;
	struct _malloc_chunk *fd;
	struct _malloc_chunk *bk;
} malloc_chunk;

#define INUSE 0x1

typedef struct {
	unsigned long desc;
	long status;
	long spec;
	long mode;
	unsigned long fhandler;
} EvCB;

#define EvStUNUSED	0x0000
#define EvStWAIT	0x1000
#define EvStACTIVE	0x2000
#define EvStALREADY 0x4000

typedef struct {
	long next;
	long func1;
	long func2;
	long pad;
} SysRPst;

static unsigned long *jmp_int = NULL;
static int *pad_buf = NULL;
static char *pad_buf1,*pad_buf2;//shadow add
static int pad_buf1len,pad_buf2len;//shadow add


static unsigned long regs[35];
static EvCB *Event;
static EvCB *HwEV; // 0xf0
static EvCB *EvEV; // 0xf1
static EvCB *RcEV; // 0xf2
static EvCB *UeEV; // 0xf3
static EvCB *SwEV; // 0xf4
static EvCB *ThEV; // 0xff
static unsigned long *heap_addr = NULL;

/*                                           *
//                                           *
//                                           *
//               System calls A0             */

void bios_atoi() { v0 = atoi((char *)Ra0); pc0 = ra;}//0x10
void bios_atol() { v0 = atoi((char *)Ra0); pc0 = ra;}//0x11
void bios_setjmp() { // 13
	unsigned long *jmp_buf= (unsigned long*)Ra0;
	int i;

	jmp_buf[0] = ra;
	jmp_buf[1] = sp;
	jmp_buf[2] = s8; // fp
	for (i=0; i<8; i++) // s0-s7
		jmp_buf[3+i] = cpuRegs.GPR.r[16+i];
	jmp_buf[11] = gp;

	v0 = 0; pc0 = ra;
}

void bios_longjmp()//14
{
	unsigned long *jmp_buf= (unsigned long*)Ra0;
	int i;
		ra = jmp_buf[0]; /* ra */
		sp = jmp_buf[1]; /* sp */
		s8 = jmp_buf[2]; /* fp */
     	for (i=0; i<8; i++) // s0-s7
		   cpuRegs.GPR.r[16+i] = jmp_buf[3+i];		
		gp = jmp_buf[11]; /* gp */
		v0 = a1; pc0 = ra;
}

/*0x15*/void bios_strcat()  { strcat(Ra0, Ra1); v0 = a0; pc0 = ra;}
/*0x16*/void bios_strncat() { strncat(Ra0, Ra1, a2); v0 = a0; pc0 = ra;}
/*0x17*/void bios_strcmp()  { v0 = strcmp(Ra0, Ra1); pc0 = ra;}
/*0x18*/void bios_strncmp() { v0 = strncmp(Ra0, Ra1, a2); pc0 = ra;}
/*0x19*/void bios_strcpy()  { strcpy(Ra0, Ra1); v0 = a0; pc0 = ra;}
/*0x1a*/void bios_strncpy() { strncpy(Ra0, Ra1, a2);  v0 = a0; pc0 = ra;}
/*0x1b*/void bios_strlen()  { v0 = strlen(Ra0); pc0 = ra;}

void bios_index() //0x1c
{ 
	char *pcA0 = (char *)Ra0; 
	char *pcRet = strchr(pcA0, a1); 
	if(pcRet) 
		v0 = a0 + pcRet - pcA0; 
	else 
		v0 = 0;
    pc0 = ra;
}
void bios_rindex() //0x1d
{ 
	char *pcA0 = (char *)Ra0; 
	char *pcRet = strrchr(pcA0, a1); 
	if(pcRet) 
		v0 = a0 + pcRet - pcA0; 
	else 
		v0 = 0;
     pc0 = ra;  
}
void bios_strchr()//0x1e
{ 
	char *pcA0 = (char *)Ra0; 
	char *pcRet = strchr(pcA0, a1); 
	if(pcRet) 
		v0 = a0 + pcRet - pcA0; 
	else 
		v0 = 0;
    pc0 = ra;
}
void bios_strrchr()//0x1f
{ 
	char *pcA0 = (char *)Ra0; 
	char *pcRet = strrchr(pcA0, a1); 
	if(pcRet) 
		v0 = a0 + pcRet - pcA0; 
	else 
		v0 = 0;
    pc0 = ra;
}
void bios_strpbrk() /*20*/ 
{ 
	char *pcA0 = (char *)Ra0; 
	char *pcRet = strpbrk(pcA0, (char *)Ra1); 
	if(pcRet) 
		v0 = a0 + pcRet - pcA0; 
	else 
		v0 = 0;
    pc0 = ra;
}

void bios_strspn()  { v0 = strspn ((char *)Ra0, (char *)Ra1); pc0 = ra;}/*21*/ 
void bios_strcspn() { v0 = strcspn((char *)Ra0, (char *)Ra1); pc0 = ra;}/*22*/ 

void bios_strtok()  /*23*/ 
{ 
	char *pcA0 = (char *)Ra0;
	char *pcRet = strtok(pcA0, (char *)Ra1);
	if(pcRet)
		v0 = a0 + pcRet - pcA0;
	else
		v0 = 0;
    pc0 = ra;
}

void bios_strstr() /*24*/  
{
	char *pcA0 = (char *)Ra0;
	char *pcRet = strstr(pcA0, (char *)Ra1);
	if(pcRet)
		v0 = a0 + pcRet - pcA0;
	else
		v0 = 0;
    pc0 = ra;
}

/*0x25*/void bios_toupper() {v0 = toupper(a0); pc0 = ra;}
/*0x26*/void bios_tolower() {v0 = tolower(a0); pc0 = ra;}
/*0x27*/void bios_bcopy()   {memcpy(Ra1,Ra0,a2); pc0=ra;}
/*0x28*/void bios_bzero()   {memset(Ra0,0,a1); pc0=ra;}
/*0x29*/void bios_bcmp()    {v0 = memcmp(Ra0,Ra1,a2); pc0=ra; }
/*0x2a*/void bios_memcpy()  {memcpy(Ra0, Ra1, a2); v0 = a0; pc0 = ra;}
/*0x2b*/void bios_memset()  {memset(Ra0, a1, a2); v0 = a0; pc0 = ra;}
/*0x2c*/void bios_memmove() {memmove(Ra0, Ra1, a2); v0 = a0; pc0 = ra;}
/*0x2d*/void bios_memcmp()  {v0 = memcmp(Ra0, Ra1, a2); pc0 = ra;}  

void bios_memchr() { // 2e
	void *ret = memchr(Ra0, a1, a2);
	if (ret != NULL) v0 = (unsigned long)((char*)ret - Ra0) + a0;
	else v0 = 0;
	pc0 = ra;
}

void bios_rand() { // 2f
	v0 = 1+(int) (32767.0*rand()/(RAND_MAX+1.0));
	pc0 = ra;
}

void bios_srand() { // 30
	srand(a0); pc0 = ra;
}

void bios_malloc() { // 33
	malloc_chunk *chunk;
	malloc_chunk *fd;

//	printf ("malloc %lx\n", a0);

	chunk = (malloc_chunk *)heap_addr;
	if (chunk == NULL) { v0 = 0; return; }

	for (; ((a0 > chunk->size) || (chunk->stat == INUSE)) || (chunk->fd != NULL); chunk = chunk->fd);
//printf ("chunk %lx\n",chunk->size);
	/* split free chunk */
	fd = chunk + sizeof(malloc_chunk) + a0;
	fd->stat = chunk->stat;
	fd->size = chunk->size - a0;
	fd->fd = chunk->fd;
	fd->bk = chunk;

	/* set new chunk */
	chunk->stat = INUSE;
	chunk->size = a0;
	chunk->fd = fd;

	v0 = ((unsigned long)chunk - (unsigned long)RAM) + sizeof(malloc_chunk);
	v0|= 0x80000000;
//	printf ("malloc %lx,%lx\n", v0, a0);
	pc0 = ra;
}

void bios_InitHeap() { // 39
	malloc_chunk *chunk;

	heap_addr = (unsigned long*)Ra0;

	chunk = (malloc_chunk *)heap_addr;
	chunk->stat = 0;
	if (((a0 & 0x1fffff) + a1)>= 0x200000) chunk->size = 0x1ffffc - (a0 & 0x1fffff);
	else chunk->size = a1;
	chunk->fd = NULL;
	chunk->bk = NULL;
//	printf ("InitHeap %lx,%lx : %lx\n",a0,a1,chunk->size);

	pc0 = ra;
}
void bios_getchar(){ v0 = getchar(); pc0=ra;} //0x3b

void bios_printf() { // 3f
	char tmp[1024];
	char tmp2[1024];
	unsigned long save[4];
	char *ptmp = tmp;
	int n=1, i=0, j;

	memcpy(save, (char*)PC_MEM(sp), 4*4);
	uRAM32(sp) = a0;
	uRAM32(sp + 4) = a1;
	uRAM32(sp + 8) = a2;
	uRAM32(sp + 12) = a3;

	while (Ra0[i]) {
		switch (Ra0[i]) {
			case '%':
				j = 0;
				tmp2[j++] = '%';
_start:
				switch (Ra0[++i]) {
					case '.':
					case 'l':
						tmp2[j++] = Ra0[i]; goto _start;
					default:
						if (Ra0[i] >= '0' && Ra0[i] <= '9') {
							tmp2[j++] = Ra0[i];
							goto _start;
						}
						break;
				}
				tmp2[j++] = Ra0[i];
				tmp2[j] = 0;

				switch (Ra0[i]) {
					case 'f': case 'F':
						ptmp+= sprintf(ptmp, tmp2, (float)uRAM32(sp + n++ * 4)); break;
					case 'a': case 'A':
					case 'e': case 'E':
					case 'g': case 'G':
						ptmp+= sprintf(ptmp, tmp2, (double)uRAM32(sp + n++ * 4)); break;
					case 'p':
					case 'i':
					case 'd': case 'D':
					case 'o': case 'O':
					case 'x': case 'X':
						ptmp+= sprintf(ptmp, tmp2, (unsigned int)uRAM32(sp + n++ * 4)); break;
					case 'c':
						ptmp+= sprintf(ptmp, tmp2, (unsigned char)uRAM32(sp + n++ * 4)); break;
					case 's':
						ptmp+= sprintf(ptmp, tmp2, (char*)PC_MEM(uRAM32(sp + n++ * 4))); break;
					case '%':
						*ptmp++ = Ra0[i]; break;
				}
				i++;
				break;
			default:
				*ptmp++ = Ra0[i++];
		}
	}
	*ptmp = 0;

	memcpy((char*)PC_MEM(sp), save, 4*4);

	SysPrintf(tmp);

	pc0 = ra;
}

void bios_FlushCache() { // 44
	pc0 = ra;
}

void bios_GPU_dw() //0x46
{

		int size;
		long *ptr;
		GPU_writeData(0xa0000000);
		GPU_writeData((a1<<16)|(a0&0xffff));
		GPU_writeData((a3<<16)|(a2&0xffff));
		size = (a2*a3+1)/2;
		ptr = (long*)Rsp[4];  //that is correct?
		do {
			GPU_writeData(*ptr++);
		} while(--size);

}  

void bios_mem2vram() //0x47
{
		int size;
		GPU_writeData(0xa0000000);
		GPU_writeData((a1<<16)|(a0&0xffff));
		GPU_writeData((a3<<16)|(a2&0xffff));
		size = (a2*a3+1)/2;
		GPU_writeStatus(0x04000002);
		hwWrite32(0x1f8010f4,0);
		hwWrite32(0x1f8010f0,hwRead32(0x1f8010f0)|0x800);
		hwWrite32(0x1f8010a0,Rsp[4]);//might have a buggy...
		hwWrite32(0x1f8010a4,((size/16)<<16)|16);
		hwWrite32(0x1f8010a8,0x01000201);
}



void bios_SendGPU() { GPU_writeStatus(a0); pc0= ra; }//0x48
void bios_GPU_cw()  { GPU_writeData(a0);   pc0 = ra;}//0x49

void bios_GPU_cwb()
{ //0x4a
	    long *ptr = (long*)Ra0;
		int size = a1;
		while(size--) 
		{
			GPU_writeData(*ptr++);
		}
}



    
void bios_GPU_SendPackets()//4b:	
{
		GPU_writeStatus(0x04000002);
		hwWrite32(0x1f8010f4,0);
		hwWrite32(0x1f8010f0,hwRead32(0x1f8010f0)|0x800);
		hwWrite32(0x1f8010a0,a0);
		hwWrite32(0x1f8010a4,0);
		hwWrite32(0x1f8010a8,0x010000401);
        pc0 = ra;

}
void bios_sys_a0_4c() //0x4c GPU relate
{
        hwWrite32(0x1f8010a8,0x00000401);
		GPU_writeData(0x0400000);
		GPU_writeData(0x0200000);
		GPU_writeData(0x0100000);
}

void bios_GPU_GetGPUStatus(){ v0 = GPU_readStatus(); pc0 = ra; }//0x4d




void bios_GetC0Table() { // 56
	pc0 = ra;
}

void bios_GetB0Table() { // 57
	pc0 = ra;
}

void bios__96_remove() { // 72
	pc0 = ra;
}

/* System calls B0 */

void bios_SetRCnt() { // 02
	a0&= 0x3;
	if (a0 != 3) {
		unsigned long mode=0;

		rcntWtarget(a0, a1);
		if (a2&0x1000) mode|= 0x050; // Interrupt Mode
		if (a2&0x0100) mode|= 0x008; // Count to 0xffff
		if (a2&0x0010) mode|= 0x001; // Timer stop mode
		if (a0 == 2) { if (a2&0x0001) mode|= 0x200; } // System Clock mode
		else         { if (a2&0x0001) mode|= 0x100; } // System Clock mode

		rcntWmode(a0, mode);
	}
	pc0 = ra;
}

void bios_GetRCnt() { // 03
	a0&= 0x3;
	if (a0 != 3) v0 = rcntRcount(a0);
	else v0 = 0;
	pc0 = ra;
}

void bios_StartRCnt() { // 04
	a0&= 0x3;
	if (a0 != 3) uHWM32(0x1074)|= (1<<(a0+4));
	else uHWM32(0x1074)|= 0x1;
	v0 = 1; pc0 = ra;
}

void bios_StopRCnt() { // 05
	a0&= 0x3;
	if (a0 != 3) uHWM32(0x1074)&= ~(1<<(a0+4));
	else uHWM32(0x1074)&= ~0x1;
	pc0 = ra;
}

void bios_ResetRCnt() { // 06
	a0&= 0x3;
	if (a0 != 3) {
		rcntWmode(a0, 0);
		rcntWtarget(a0, 0);
		rcntWcount(a0, 0);
	}
	pc0 = ra;
}


/* gets ev for use with Event */
#define GetEv() \
	ev = a0>>28; \
	if (ev == 0xf) ev = 0x5; \
	ev*= 32; \
	ev+= a0&0x1f;

void bios_DeliverEvent() { // 07
	int ev;

	GetEv();

	pc0 = ra;
}

void bios_OpenEvent() { // 08
	int ev;

//	printf ("syscall b0:08; OpenEvent (class:%lx, spec:%lx, mode:%lx, func:%lx)\n", a0, a1, a2, a3);

	GetEv();

	Event[ev].status = EvStWAIT;
	Event[ev].mode = a2;
	Event[ev].fhandler = a3;

	v0 = a0;
	pc0 = ra;
}

void bios_CloseEvent() { // 09
	int ev;

	GetEv();

	Event[ev].status = EvStUNUSED;

	v0 = 1; pc0 = ra;
}

void bios_WaitEvent() { // 0a
	int ev;

	GetEv();

//	printf ("syscall WaitEvent %lx\n", a0);

	Event[ev].status = EvStACTIVE;

	v0 = 1; pc0 = ra;
}

void bios_TestEvent() { // 0b
	int ev;

	GetEv();

	if (Event[ev].status == EvStALREADY) { Event[ev].status = EvStACTIVE; v0 = 1; }
	else v0 = 0;
v0=1;
	pc0 = ra;
}

void bios_EnableEvent() { // 0c
	int ev;

	GetEv();

	Event[ev].status = EvStACTIVE;

	v0 = 1; pc0 = ra;
}

void bios_DisableEvent() { // 0d
	int ev;

	GetEv();

	Event[ev].status = EvStWAIT;

	v0 = 1; pc0 = ra;
}

void bios_InitPAD()//0x12  don't know about this (shadow)
{
		pad_buf1 = (char*)Ra0;
		pad_buf1len = a1;
		pad_buf2 = (char*)Ra2;
		pad_buf2len = a3;

}


void bios_StartPAD()
{//13
	hwWrite16(0x1f801074, (unsigned short)(hwRead16(0x1f801074) | 0x1));
	cpuRegs.CP0.n.Status |= 0x401;
	pc0 = ra;
}


void bios_PAD_init() { // 15
	hwWrite16(0x1f801074, (unsigned short)(hwRead16(0x1f801074) | 0x1));
	pad_buf = (int*)Ra1;
	*pad_buf = -1;
	cpuRegs.CP0.n.Status |= 0x401;
	pc0 = ra;
}

void bios_PAD_dr() { // 16
	v0 = -1;
	pc0 = ra;
}

void bios_ReturnFromException() { // 17
	memcpy(cpuRegs.GPR.r, regs, 32*4);
	cpuRegs.GPR.n.lo = regs[32];
	cpuRegs.GPR.n.hi = regs[33];
	cpuRegs.pc = regs[34];

	pc0 = cpuRegs.CP0.n.EPC;
	if (cpuRegs.CP0.n.Cause & 0x80000000) pc0+=4;

	cpuRegs.CP0.n.Status = (cpuRegs.CP0.n.Status & 0xfffffff0) |
						  ((cpuRegs.CP0.n.Status & 0x3c) >> 2);
}


void bios_HookEntryInt() { // 18
	jmp_int = (unsigned long*)Ra0;
	pc0 = ra;
}
void bios_ResetEntryInt() { // 19
	jmp_int = NULL;
	pc0 = ra;
}



    
void bios_write() { // 35/03
    if (a0 == 1) {
		SysPrintf(Ra1);
    }
	pc0 = ra;
}

void bios_putchar () { // 3d
    char tmp[12];

    sprintf (tmp,"%c",(char)a0);
    SysPrintf(tmp);

    pc0 = ra;
}

void bios_puts () { // 3e/3f
    SysPrintf(Ra0);

    pc0 = ra;
}

void bios_ChangeClearPad() { // 5b
	pc0 = ra;
}

/* System calls C0 */

void bios_SysEnqIntRP() { // 02
	
	pc0 = ra;
}

void bios_SysDeqIntRP() { // 03
	pc0 = ra;
}

void bios_ChangeClearRCnt() { // 0a
	pc0 = ra;
}

void bios_dummy() 
{ 
#ifdef BIOS_LOG
	BIOS_LOG("unk %lx call: %lx\n", pc0 & 0x1fffff, t1);
#endif
//	printf("unk %lx call: %lx\n", pc0 & 0x1fffff, t1);
	pc0 = ra; 

}

void (*biosA0[256])();
void (*biosB0[256])();
void (*biosC0[256])();

void biosInit() {
	unsigned long size;
	int i;

	for(i = 0; i < 256; i++) {
		biosA0[i] = NULL;
		biosB0[i] = NULL;
		biosC0[i] = NULL;
	}
	biosA0[0x3e] = bios_puts;
	biosA0[0x3f] = bios_printf;

	biosB0[0x3d] = bios_putchar;
	biosB0[0x3f] = bios_puts;

	if (!Config.HLE) return;

	for(i = 0; i < 256; i++) {
		if (biosA0[i] == NULL) biosA0[i] = bios_dummy;
		if (biosB0[i] == NULL) biosB0[i] = bios_dummy;
		if (biosC0[i] == NULL) biosC0[i] = bios_dummy;
	}
    //biosA0[0x00] = bios_open;
	//biosA0[0x01] = bios_lseek;
	//biosA0[0x02] = bios_read;
	biosA0[0x03] = bios_write;
	//biosA0[0x04] = bios_close;
	//biosA0[0x05] = bios_ioctl;
	//biosA0[0x06] = bios_exit;
	//biosA0[0x07] = bios_sys_a0_07;
	//biosA0[0x08] = bios_getc;
	//biosA0[0x09] = bios_putc;
	//biosA0[0x0a] = bios_todigit;
	//biosA0[0x0b] = bios_atof;
	//biosA0[0x0c] = bios_strtoul;
	//biosA0[0x0d] = bios_strtol;
	//biosA0[0x0e] = bios_abs;
	//biosA0[0x0f] = bios_labs;
    biosA0[0x10] = bios_atoi;  
    biosA0[0x11] = bios_atol;  
	//biosA0[0x12] = bios_atob;
	biosA0[0x13] = bios_setjmp;
	biosA0[0x14] = bios_longjmp;
	biosA0[0x15] = bios_strcat;
	biosA0[0x16] = bios_strncat;
	biosA0[0x17] = bios_strcmp;
	biosA0[0x18] = bios_strncmp;
	biosA0[0x19] = bios_strcpy;
	biosA0[0x1a] = bios_strncpy;
	biosA0[0x1b] = bios_strlen;
	biosA0[0x1c] = bios_index;
	biosA0[0x1d] = bios_rindex;
	biosA0[0x1e] = bios_strchr;
	biosA0[0x1f] = bios_strrchr;
	biosA0[0x20] = bios_strpbrk;
	biosA0[0x21] = bios_strspn;
	biosA0[0x22] = bios_strcspn;
	biosA0[0x23] = bios_strtok;
	biosA0[0x24] = bios_strstr;
    biosA0[0x25] = bios_toupper;
    biosA0[0x26] = bios_tolower;
    biosA0[0x27] = bios_bcopy;
    biosA0[0x28] = bios_bzero;
    biosA0[0x29] = bios_bcmp;
	biosA0[0x2a] = bios_memcpy;
	biosA0[0x2b] = bios_memset;
	biosA0[0x2c] = bios_memmove;
	biosA0[0x2d] = bios_memcmp;
	biosA0[0x2e] = bios_memchr;
	biosA0[0x2f] = bios_rand;
	biosA0[0x30] = bios_srand;
	//biosA0[0x31] = bios_qsort;
	//biosA0[0x32] = bios_strtod;
	biosA0[0x33] = bios_malloc;
	//biosA0[0x34] = bios_free;
    //biosA0[0x35] = bios_lsearch;
    //biosA0[0x36] = bios_bsearch;
    //biosA0[0x37] = bios_calloc;
    //biosA0[0x38] = bios_realloc;
	biosA0[0x39] = bios_InitHeap;
    //biosA0[0x3a] = bios__exit;
	biosA0[0x3b] = bios_getchar;
	biosA0[0x3c] = bios_putchar;	
	//biosA0[0x3d] = bios_gets;		
	//biosA0[0x40] = bios_sys_a0_40;		
	//biosA0[0x41] = bios_LoadTest;				
	//biosA0[0x42] = bios_Load;		
	//biosA0[0x43] = bios_Exec;
	biosA0[0x44] = bios_FlushCache;
    //biosA0[0x45] = bios_InstallInterruptHandler;
	biosA0[0x46] = bios_GPU_dw;
	biosA0[0x47] = bios_mem2vram;
	biosA0[0x48] = bios_SendGPU; 
	biosA0[0x49] = bios_GPU_cw;
    biosA0[0x4a] = bios_GPU_cwb;
	biosA0[0x4b] = bios_GPU_SendPackets;
    biosA0[0x4c] = bios_sys_a0_4c;
	biosA0[0x4d] = bios_GPU_GetGPUStatus;
	//biosA0[0x4e] = bios_GPU_sync;	
	//biosA0[0x4f] = bios_sys_a0_4f;
	//biosA0[0x50] = bios_sys_a0_50;		
	//biosA0[0x51] = bios_LoadExec;				
	//biosA0[0x52] = bios_GetSysSp;		
	//biosA0[0x53] = bios_sys_a0_53;
	//biosA0[0x54] = bios__96_init_a54;
	//biosA0[0x55] = bios__bu_init_a55;			
	//biosA0[0x56] = bios__96_remove_a56;	
	//biosA0[0x57] = bios_sys_a0_57;
	//biosA0[0x58] = bios_sys_a0_58;		
	//biosA0[0x59] = bios_sys_a0_59;			
	//biosA0[0x5a] = bios_sys_a0_5a;	
	//biosA0[0x5b] = bios_dev_tty_init;
	//biosA0[0x5c] = bios_dev_tty_open;		
	//biosA0[0x5d] = bios_sys_a0_5d;			
	//biosA0[0x5e] = bios_dev_tty_ioctl;
	//biosA0[0x5f] = bios_dev_cd_open;
	//biosA0[0x60] = bios_dev_cd_read;
	//biosA0[0x61] = bios_dev_cd_close;
	//biosA0[0x62] = bios_dev_cd_firstfile;
	//biosA0[0x63] = bios_dev_cd_nextfile;
	//biosA0[0x64] = bios_dev_cd_chdir;
	//biosA0[0x65] = bios_dev_card_open;
	//biosA0[0x66] = bios_dev_card_read;
	//biosA0[0x67] = bios_dev_card_write;
	//biosA0[0x68] = bios_dev_card_close;
	//biosA0[0x69] = bios_dev_card_firstfile;
	//biosA0[0x6a] = bios_dev_card_nextfile;
	//biosA0[0x6b] = bios_dev_card_erase;
	//biosA0[0x6c] = bios_dev_card_undelete;
	//biosA0[0x6d] = bios_dev_card_format;
	//biosA0[0x6e] = bios_dev_card_rename;
	//biosA0[0x6f] = bios_dev_card_6f;
	//biosA0[0x70] = bios__bu_init;
	//biosA0[0x71] = bios__96_init;
	biosA0[0x72] = bios__96_remove;
	//biosA0[0x73] = bios_sys_a0_73;
	//biosA0[0x74] = bios_sys_a0_74;
	//biosA0[0x75] = bios_sys_a0_75;
	//biosA0[0x76] = bios_sys_a0_76;
	//biosA0[0x77] = bios_sys_a0_77;
	//biosA0[0x78] = bios__96_CdSeekL;
	//biosA0[0x79] = bios_sys_a0_79;
	//biosA0[0x7a] = bios_sys_a0_7a;
	//biosA0[0x7b] = bios_sys_a0_7b;
	//biosA0[0x7c] = bios__96_CdGetStatus;
	//biosA0[0x7d] = bios_sys_a0_7d;
	//biosA0[0x7e] = bios__96_CdRead;
	//biosA0[0x7f] = bios_sys_a0_7f;
	//biosA0[0x80] = bios_sys_a0_80;
	//biosA0[0x81] = bios_sys_a0_81;
	//biosA0[0x82] = bios_sys_a0_82;		
	//biosA0[0x83] = bios_sys_a0_83;
	//biosA0[0x84] = bios_sys_a0_84;
	//biosA0[0x85] = bios__96_CdStop;	
	//biosA0[0x86] = bios_sys_a0_86;
	//biosA0[0x87] = bios_sys_a0_87;
	//biosA0[0x88] = bios_sys_a0_88;
	//biosA0[0x89] = bios_sys_a0_89;
	//biosA0[0x8a] = bios_sys_a0_8a;
	//biosA0[0x8b] = bios_sys_a0_8b;
	//biosA0[0x8c] = bios_sys_a0_8c;
	//biosA0[0x8d] = bios_sys_a0_8d;
	//biosA0[0x8e] = bios_sys_a0_8e;		
	//biosA0[0x8f] = bios_sys_a0_8f;
	//biosA0[0x90] = bios_sys_a0_90;
	//biosA0[0x91] = bios_sys_a0_91;
	//biosA0[0x92] = bios_sys_a0_92;
	//biosA0[0x93] = bios_sys_a0_93;
	//biosA0[0x94] = bios_sys_a0_94;
	//biosA0[0x95] = bios_sys_a0_95;
	//biosA0[0x96] = bios_AddCDROMDevice;
	//biosA0[0x97] = bios_AddMemCardDevide;
	//biosA0[0x98] = bios_DisableKernelIORedirection;
	//biosA0[0x99] = bios_EnableKernelIORedirection;
	//biosA0[0x9a] = bios_sys_a0_9a;
	//biosA0[0x9b] = bios_sys_a0_9b;
	//biosA0[0x9c] = bios_SetConf;
	//biosA0[0x9d] = bios_GetConf;
	//biosA0[0x9e] = bios_sys_a0_9e;
	//biosA0[0x9f] = bios_SetMem;
	//biosA0[0xa0] = bios__boot;
	//biosA0[0xa1] = bios_SystemError;
	//biosA0[0xa2] = bios_EnqueueCdIntr;
	//biosA0[0xa3] = bios_DequeueCdIntr;
	//biosA0[0xa4] = bios_sys_a0_a4;
	//biosA0[0xa5] = bios_ReadSector;
	//biosA0[0xa6] = bios_get_cd_status;
	//biosA0[0xa7] = bios_bufs_cb_0;
	//biosA0[0xa8] = bios_bufs_cb_1;
	//biosA0[0xa9] = bios_bufs_cb_2;
	//biosA0[0xaa] = bios_bufs_cb_3;
	//biosA0[0xab] = bios__card_info;
	//biosA0[0xac] = bios__card_load;
	//biosA0[0axd] = bios__card_auto;
	//biosA0[0xae] = bios_bufs_cd_4;
	//biosA0[0xaf] = bios_sys_a0_af;
	//biosA0[0xb0] = bios_sys_a0_b0;
	//biosA0[0xb1] = bios_sys_a0_b1;
	//biosA0[0xb2] = bios_do_a_long_jmp
	//biosA0[0xb3] = bios_sys_a0_b3;
	//biosA0[0xb4] = bios_sub_function;
//*******************B0 CALLS****************************
	//biosB0[0x00] = bios_SysMalloc;
	//biosB0[0x01] = bios_sys_b0_01;
	biosB0[0x02] = bios_SetRCnt;
	biosB0[0x03] = bios_GetRCnt;
	biosB0[0x04] = bios_StartRCnt;
	biosB0[0x05] = bios_StopRCnt;
	biosB0[0x06] = bios_ResetRCnt;
	biosB0[0x07] = bios_DeliverEvent;
	biosB0[0x08] = bios_OpenEvent;
	biosB0[0x09] = bios_CloseEvent;
	biosB0[0x0a] = bios_WaitEvent;
	biosB0[0x0b] = bios_TestEvent;
	biosB0[0x0c] = bios_EnableEvent;
	biosB0[0x0d] = bios_DisableEvent;
    //biosB0[0x0e] = bios_OpenTh;
    //biosB0[0x0f] = bios_CloseTh;
    //biosB0[0x10] = bios_ChangeTh;
    //biosB0[0x11] = bios_bios_b0_11;
	biosB0[0x12] = bios_InitPAD;
	biosB0[0x13] = bios_StartPAD;
	//biosB0[0x14] = bios_StopPAD;
	biosB0[0x15] = bios_PAD_init;
	biosB0[0x16] = bios_PAD_dr;
	biosB0[0x17] = bios_ReturnFromException;
	biosB0[0x18] = bios_ResetEntryInt;
	biosB0[0x19] = bios_HookEntryInt;
    //biosB0[0x1a] = bios_sys_b0_1a;
	//biosB0[0x1b] = bios_sys_b0_1b;
	//biosB0[0x1c] = bios_sys_b0_1c;
	//biosB0[0x1d] = bios_sys_b0_1d;
	//biosB0[0x1e] = bios_sys_b0_1e;
	//biosB0[0x1f] = bios_sys_b0_1f;
	//biosB0[0x20] = bios_UnDeliverEvent;
	//biosB0[0x21] = bios_sys_b0_21;
	//biosB0[0x22] = bios_sys_b0_22;
	//biosB0[0x23] = bios_sys_b0_23;
	//biosB0[0x24] = bios_sys_b0_24;
	//biosB0[0x25] = bios_sys_b0_25;
	//biosB0[0x26] = bios_sys_b0_26;
	//biosB0[0x27] = bios_sys_b0_27;
	//biosB0[0x28] = bios_sys_b0_28;
	//biosB0[0x29] = bios_sys_b0_29;
	//biosB0[0x2a] = bios_sys_b0_2a;
	//biosB0[0x2b] = bios_sys_b0_2b;
	//biosB0[0x2c] = bios_sys_b0_2c;
	//biosB0[0x2d] = bios_sys_b0_2d;
	//biosB0[0x2e] = bios_sys_b0_2e;
	//biosB0[0x2f] = bios_sys_b0_2f;
	//biosB0[0x30] = bios_sys_b0_30;
	//biosB0[0x31] = bios_sys_b0_31;
    //biosB0[0x32] = bios_open;
	//biosB0[0x33] = bios_lseek;
	//biosB0[0x34] = bios_read;		
	biosB0[0x35] = bios_write;
    //biosB0[0x36] = bios_close;
	//biosB0[0x37] = bios_ioctl;
	//biosB0[0x38] = bios_exit;
	//biosB0[0x39] = bios_sys_b0_39;
	//biosB0[0x3a] = bios_getc;
	//biosB0[0x3b] = bios_putc;
	biosB0[0x3c] = bios_getchar;
	//biosB0[0x3e] = bios_gets;
	//biosB0[0x40] = bios_cd;
	//biosB0[0x41] = bios_format;
	//biosB0[0x42] = bios_firstfile;
	//biosB0[0x43] = bios_nextfile;
	//biosB0[0x44] = bios_rename;
	//biosB0[0x45] = bios_delete;
	//biosB0[0x46] = bios_undelete;
	//biosB0[0x47] = bios_AddDevice;
	//biosB0[0x48] = bios_RemoteDevice;
	//biosB0[0x49] = bios_PrintInstalledDevices;
	//biosB0[0x4a] = bios_InitCARD;
	//biosB0[0x4b] = bios_StartCARD;
	//biosB0[0x4c] = bios_StopCARD;
	//biosB0[0x4d] = bios_sys_b0_4d;
	//biosB0[0x4e] = bios__card_write;
	//biosB0[0x4f] = bios__card_read;
	//biosB0[0x50] = bios__new_card;
	//biosB0[0x51] = bios_Krom2RawAdd;
	//biosB0[0x52] = bios_sys_b0_52;
	//biosB0[0x53] = bios_sys_b0_53;
	//biosB0[0x54] = bios__get_errno;
	//biosB0[0x55] = bios__get_error;
	biosB0[0x56] = bios_GetC0Table;
	biosB0[0x57] = bios_GetB0Table;
    //biosB0[0x58] = bios__card_chan;
	//biosB0[0x59] = bios_sys_b0_59;
	//biosB0[0x5a] = bios_sys_b0_5a;
	biosB0[0x5b] = bios_ChangeClearPad;
	//biosB0[0x5c] = bios__card_status;
	//biosB0[0x5d] = bios__card_wait;
//*******************C0 CALLS****************************
	//biosC0[0x00] = bios_InitRCnt;
	//biosC0[0x01] = bios_InitException;
	biosC0[0x02] = bios_SysEnqIntRP;
	biosC0[0x03] = bios_SysDeqIntRP;
	//biosC0[0x04] = bios_get_free_EvCB_slot;
	//biosC0[0x05] = bios_get_free_TCB_slot;
	//biosC0[0x06] = bios_ExceptionHandler;
	//biosC0[0x07] = bios_InstallExeptionHandler;
	//biosC0[0x08] = bios_SysInitMemory;
	//biosC0[0x09] = bios_SysInitKMem;
	biosC0[0x0a] = bios_ChangeClearRCnt;	
	//biosC0[0x0b] = bios_SystemError;
	//biosC0[0x0c] = bios_InitDefInt;
    //biosC0[0x0d] = bios_sys_c0_0d;
	//biosC0[0x0e] = bios_sys_c0_0e;
	//biosC0[0x0f] = bios_sys_c0_0f;
	//biosC0[0x10] = bios_sys_c0_10;
	//biosC0[0x11] = bios_sys_c0_11;
	//biosC0[0x12] = bios_InstallDevices;
	//biosC0[0x13] = bios_FlushStfInOutPut;
	//biosC0[0x14] = bios_sys_c0_14;
	//biosC0[0x15] = bios__cdevinput;
	//biosC0[0x16] = bios__cdevscan;
	//biosC0[0x17] = bios__circgetc;
	//biosC0[0x18] = bios__circputc;		  
	//biosC0[0x19] = bios_ioabort;
	//biosC0[0x1a] = bios_sys_c0_1a
	//biosC0[0x1b] = bios_KernelRedirect;
	//biosC0[0x1c] = bios_PatchAOTable;
//************** THE END ***************************************




	size = sizeof(EvCB);
	Event = malloc(size * 6);
	if (Event == NULL) { SysMessage("Error allocating memory"); SysClose(); exit(1); }
	HwEV = Event;
	EvEV = Event + size;
	RcEV = Event + size*2;
	UeEV = Event + size*3;
	SwEV = Event + size*4;
	ThEV = Event + size*5;
}

void biosShutdown() {
	free(Event);
}

static unsigned long softPC0, pra;
static int softInt = 0;

void biosInterrupt() {
	if ((cpuRegs.CP0.n.Status & 0x404) != 0x404) return;

	if (cpuRegs.CP0.n.Cause & 0x100) {
		cpuRegs.CP0.n.Cause&=~0x100;
		ra = pra;
		return;
	}

	if (uHWM32(0x1070) & 0x1) 
	{ // Vsync
		if (pad_buf) 
		{
			PadDataS pad;

			PAD1_readPort1(&pad);
			*pad_buf = ((pad.buttonStatus & 0xff) << 8) | (pad.buttonStatus >> 8);
			PAD2_readPort2(&pad);
			*pad_buf|=(((pad.buttonStatus & 0xff) << 8) | (pad.buttonStatus >> 8)) << 16;
		}
        if (pad_buf1) 
		{
	       PadDataS pad;
	       pad_buf1[0] = 0;
		   pad_buf1[1] = 0x41;
           PAD1_readPort1(&pad);
		   pad_buf1[2] = ((pad.buttonStatus & 0xff) << 8) | (pad.buttonStatus >> 8) >> 8;
		   pad_buf1[3] = ((pad.buttonStatus & 0xff) << 8) | (pad.buttonStatus >> 8);
		   pad_buf2[0] = 0;
		   pad_buf2[1] = 0x41;
           PAD2_readPort2(&pad);
		   pad_buf2[2] = (((pad.buttonStatus & 0xff) << 8) | (pad.buttonStatus >> 8))>>8;
		   pad_buf2[3] = (((pad.buttonStatus & 0xff) << 8) | (pad.buttonStatus >> 8));
		}


	}

	if (uHWM32(0x1070) & 0x70) { // Rcnt 0,1,2
		int i;

		for (i=0; i<3; i++) {
			if (counters[i].mode & 0x50) {
				if (RcEV[i].status == EvStACTIVE) {
//					printf ("rcnt int\n");
					softInt = 1;
					softPC0 = RcEV[i].fhandler;
					return;
				}
			}
		}
	}

//	cpuRegs.CP0.n.Status &= ~0x4;
}

void biosException() {
	switch ((cpuRegs.CP0.n.Cause & 0x3c) >> 2) {
		case 0: // Interrupt
#ifdef CPU_LOG
			CPU_LOG("interrupt\n");
#endif
			memcpy(regs, cpuRegs.GPR.r, 32*4);
			regs[32] = cpuRegs.GPR.n.lo;
			regs[33] = cpuRegs.GPR.n.hi;
			regs[34] = cpuRegs.pc;

			biosInterrupt();

			if (softInt) {
				cpuRegs.CP0.n.Cause|= 0x100;
				pc0 = softPC0;
				pra = ra;
				ra = 0x00000080;
				softInt = 0;
//				printf("soft int\n");
				return;
			}

			if (jmp_int != NULL) {
				int i;

				ra = jmp_int[0];
				sp = jmp_int[1];
				s8 = jmp_int[2]; // fp
				for (i=0; i<8; i++) // s0-s7
					 cpuRegs.GPR.r[16+i] = jmp_int[3+i];
				gp = jmp_int[11];

				v0 = 1;
				pc0 = ra;
				hwWrite16(0x1f801070, 0xffff);
				return;
			}
			else hwWrite16(0x1f801070, 0);
			break;
		case 8: // Syscall
#ifdef CPU_LOG
			CPU_LOG("syscall exp %x\n", a0);
#endif
			switch (a0) {
				case 1: // EnterCritical - disable irq's
					cpuRegs.CP0.n.Status&=~0x404; break;
				case 2: // ExitCritical - enable irq's
					cpuRegs.CP0.n.Status|= 0x404; break;
			}
			pc0 = cpuRegs.CP0.n.EPC + 4;

			cpuRegs.CP0.n.Status = (cpuRegs.CP0.n.Status & 0xfffffff0) |
								  ((cpuRegs.CP0.n.Status & 0x3c) >> 2);
			return;
		default:
#ifdef CPU_LOG
			CPU_LOG("unk exp\n");
#endif
			break;
	}

	pc0 = cpuRegs.CP0.n.EPC;
	if (cpuRegs.CP0.n.Cause & 0x80000000) pc0+=4;

	cpuRegs.CP0.n.Status = (cpuRegs.CP0.n.Status & 0xfffffff0) |
						  ((cpuRegs.CP0.n.Status & 0x3c) >> 2);
}
