/*  Pcsx - Pc Psx Emulator
 *  Copyright (C) 1999-2002  Pcsx Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <string.h>

#include "Common.h"

int cnts = 4;

void rcntUpd(unsigned long index) {
	counters[index].sCycle = cpuIrq.cycle;
	if ((!(counters[index].mode & 1)) || (index!=2)) {
		if (counters[index].mode & 0x08) { // Count to target
			counters[index].Cycle = ((counters[index].target - counters[index].count) * counters[index].rate) / BIAS;
		}
		else { // Count to 0xffff
			counters[index].Cycle = ((0xffff - counters[index].count) * counters[index].rate) / BIAS;
		}
	} else counters[index].Cycle = 0xffffffff;
}

void rcntReset(unsigned long index) {
	counters[index].count = 0;
	rcntUpd(index);

	if ((counters[index].mode & 0x50) == 0x50) uHWM32(0x1070)|= counters[index].interrupt;
}

void rcntSet() {
	int i;

	nextCounter = 0x7fffffff;
	nextsCounter = cpuIrq.cycle;

	for (i = 0; i < cnts; i++) {
		long count = counters[i].Cycle - (cpuIrq.cycle - counters[i].sCycle);

		if (count < 0) {
			nextCounter = 0; break;
		}

		if (count < (long)nextCounter) {
			nextCounter = count;
		}
	}
}

void rcntInit() {
	nextCounter = 0;

	memset(counters, 0, sizeof(counters));

	counters[0].rate = 1; counters[0].interrupt = 16;
	counters[1].rate = 1; counters[1].interrupt = 32;
	counters[2].rate = 1; counters[2].interrupt = 64;

	counters[3].interrupt = 1;
	counters[3].mode = 0x58; // The VSync counter mode
	counters[3].target = 1;
	UpdateVSyncRate();

	if (Config.PsxType)
		counters[4].rate = (262 / 32) * 50; // pal
	else
		counters[4].rate = (312 / 32) * 60; // ntsc
	counters[4].target = 1;
	counters[4].mode = 0x8;
	if (SPU_update != NULL) cnts = 5;

	rcntUpd(0); rcntUpd(1); rcntUpd(2); rcntUpd(3);
	rcntSet();
}

void UpdateVSyncRate() {
	if (Config.PsxType) // ntsc - 0 | pal - 1
	     counters[3].rate = (PSXCLK / 50);// / BIAS;
	else counters[3].rate = (PSXCLK / 60);// / BIAS;
}

void rcntUpdate() {
	if ((cpuIrq.cycle - counters[3].sCycle) >= counters[3].Cycle) {
		rcntReset(3);
		GPU_updateLace(); // updateGPU
		SysUpdate();
#ifdef GTE_LOG
		GTE_LOG("VSync\n");
#endif
	}

	if ((cpuIrq.cycle - counters[0].sCycle) >= counters[0].Cycle) {
		rcntReset(0);
	}

	if ((cpuIrq.cycle - counters[1].sCycle) >= counters[1].Cycle) {
		rcntReset(1);
	}

	if ((cpuIrq.cycle - counters[2].sCycle) >= counters[2].Cycle) {
		rcntReset(2);
	}

	if (SPU_update != NULL) {
		if ((cpuIrq.cycle - counters[4].sCycle) >= counters[4].Cycle) {
			rcntReset(4);
			SPU_update();
		}
	}

	rcntSet();
}

void rcntWcount(unsigned long index, unsigned long value) {
//	SysPrintf ("writeCcount[%d] = %x\n", index, value);
//	CPU_LOG("writeCcount[%d] = %x\n", index, value);
	counters[index].count = value;
	rcntUpd(index);
	rcntSet();
}

void rcntWmode(unsigned long index, unsigned long value)  {
//	SysPrintf ("writeCmode[%ld] = %lx\n", index, value);
	counters[index].mode = value;
	if(index == 0) {
		switch (value & 0x300) {
			case 0x100:
				counters[index].rate = (counters[3].rate /** BIAS*/) / 386 / 262; // seems ok
				break;
			default:
				counters[index].rate = 1;
		}
	}
	else if(index == 1) {
		switch (value & 0x300) {
			case 0x100:
				counters[index].rate = (counters[3].rate /** BIAS*/) / 262; // seems ok
				break;
			default:
				counters[index].rate = 1;
		}
	}
	else if(index == 2) {
		switch (value & 0x300) {
			case 0x200:
				counters[index].rate = 8; // 1/8 speed
				break;
			default:
				counters[index].rate = 1; // normal speed
		}
	}

	// Need to set a rate and target
	rcntUpd(index);
	rcntSet();
}

void rcntWtarget(unsigned long index, unsigned long value) {
//	SysPrintf ("writeCtarget[%ld] = %lx\n", index, value);
	counters[index].target = value;
	rcntUpd(index);
	rcntSet();
}

unsigned long rcntRcount(unsigned long index) {
//	if ((!(counters[index].mode & 1)) || (index!=2)) {
		return (counters[index].count + BIAS * (cpuIrq.cycle - counters[index].sCycle) / counters[index].rate) & 0xffff;
//	} else return 0;
}

int rcntFreeze(gzFile f, int Mode) {
	char Unused[4096 - sizeof(Counter)];

	gzfreezel(counters);
	gzfreezel(Unused);

	return 0;
}
