/*  Pcsx - Pc Psx Emulator
 *  Copyright (C) 1999-2002  Pcsx Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef __WIN32__
#pragma warning(disable:4244)
#endif
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "R3000A.h"
#include "Common.h"
#include "ix86.h"

//#define CACHE_REGS

unsigned long *recLUT;

#define PC_REC(x) (recLUT[x >> 16] + (x & 0xffff))
#define PC_REC8(x) (*(unsigned char *)PC_REC(x))
#define PC_REC16(x) (*(unsigned short *)PC_REC(x))
#define PC_REC32(x) (*(unsigned long *)PC_REC(x))

static char *recMem;				/* the recompiled blocks will be here */
static char *recRAM;				/* and the ptr to the blocks here */
static char *recROM;				/* and here */
static char *recPtr;				/* pointer to recMem */

static unsigned long pc;			/* recompiler pc */
static int count;					/* recompiler intruction count */
static int branch;					/* set for branch */
static unsigned long target;		/* branch target */

static int mode;

static void (*recBSC[64])();
static void (*recSPC[64])();
static void (*recREG[32])();
static void (*recCP0[32])();
static void (*recCP2[64])();
static void (*recCP2BSC[32])();

#ifdef CACHE_REGS

static unsigned long regs[32];		/* 32 gpregs lookup table */
static unsigned long regsmode[32];	/* 32 gpregs mode table */
static unsigned long regsm;			/* 32 gpregs mask */
static unsigned long x86regs[8];	/* 8 x86 lookup table */
static unsigned long x86regsc;		/* 8 x86 regs count */
static unsigned long x86regsm;		/* 8 x86 regs mask */
static unsigned long x86skip;		/* set to skip on CheckReg */

/* save a x86 register into the R3000A register */
#define SaveReg(r3kreg, x86reg) {\
	MOV32RtoM((unsigned long)&cpuRegs.GPR.r[r3kreg], x86reg); }

/* loads a R3000A register into a x86 register */
#define LoadReg(x86reg, r3kreg) { \
	MOV32MtoR(x86reg, (unsigned long)&cpuRegs.GPR.r[r3kreg]); }

/* removes R3000A register from x86 reg, if set */
#define RmoveReg(reg) \
	if (regsm & (1<<reg)) { \
		if (regsmode[reg] == RM_W) \
			SaveReg(reg, regs[reg]); \
		x86regsm &=~(1<<regs[reg]); \
		regsm &= ~(1<<reg); \
	}

/* clear R3000A register from x86 reg, if set */
#define ClearX86Reg(reg) \
	if (x86regsm & (1<<reg)) { \
		if (regsmode[x86regs[reg]] == RM_W) \
			SaveReg(x86regs[reg], reg); \
		x86regsm &=~(1<<reg); \
		regsm &=~(1<<x86regs[reg]); /* clear old (r3k)reg from (r3k)regmask */ \
	}


#define RM_R	0x1 // read mode
#define RM_W	0x2 // write mode

/* this checks a R3000A register to use and uploads to the next avail x86 reg */
void CheckReg(unsigned long reg, int mode) {
	if (!(regsm & (1<<reg))) {
		if (x86regsm & (1<<x86regsc)) {
//printf ("save %d mode %d\n",reg,regsmode[reg]);
			if (regsmode[x86regs[x86regsc]] == RM_W)
				SaveReg(x86regs[x86regsc], x86regsc);
			regsm &=~(1<<x86regs[x86regsc]); /* clear old (r3k)reg from (r3k)regmask */
		} else {
			x86regsm |= (1<<x86regsc); /* set x86regs mask with x86reg */
		}
//printf ("load %d mode %d (in %d)\n",reg,mode,x86regsc);
		x86regs[x86regsc] = reg; /* set x86regs with new (r3k)reg */
		regs[reg] = x86regsc; /* set (r3k)regs with x86reg */
		LoadReg(regs[reg], reg);
		regsmode[reg] = mode;
		regsm|= (1<<reg);  /* or new (r3k)reg to (r3k)regmask */
		x86regsc++; /* increment x86regsc */
		while ((x86regsc == EAX) || (x86regsc == ESP) || (x86regsc == x86skip))
			x86regsc++; /* skip regs */
		if (x86regsc >= 8) x86regsc = (x86skip == ECX) ? EDX : ECX;
	} else {
//printf ("check %d mode %d\n",reg,mode);
		if (mode == RM_W) regsmode[reg] = RM_W;
	}
}

/* updates all R3000A regs */
void FlushRegs() {
	int i;

	for (i=1; i<8; i++) {
		ClearX86Reg(i);
	}
}

#else

void FlushRegs() {}

#endif

/* set a pending branch */
#define SetBranch() { \
	/*FlushRegs();*/ \
	branch = 1; \
	cpuRegs.code = PC_MEM32(pc); \
	pc+=4; \
	count++; \
/*	MOV32ItoR(EAX, (unsigned long)pc);*/ \
/*	MOV32RtoM((unsigned long)&cpuRegs.pc, EAX);*/ \
	recBSC[cpuRegs.code>>26](); \
	FlushRegs(); \
/*	PUSHA32();*/ \
	MOV32MtoR(EAX, (unsigned long)&target); \
	MOV32RtoM((unsigned long)&cpuRegs.pc, EAX); \
	MOV32ItoR(EAX, (unsigned long)cpuBranchTest); \
	CALL32R  (EAX); \
/*	POPA32();*/ \
}

#define REC_FUNC(f) \
void int##f(); \
void rec##f() { \
	FlushRegs(); \
 \
	MOV32ItoR(EAX, (unsigned long)cpuRegs.code); \
	MOV32RtoM((unsigned long)&cpuRegs.code, EAX); \
	MOV32ItoR(EAX, (unsigned long)pc); \
	MOV32RtoM((unsigned long)&cpuRegs.pc, EAX); \
	MOV32ItoR(EAX, (unsigned long)int##f); \
	CALL32R  (EAX); \
	MOV32ItoR(EAX, 0); \
	MOV32RtoM((unsigned long)&cpuRegs.GPR.n.r0, EAX); \
/*	branch = 2; */\
}

#define REC_SYS(f) \
void int##f(); \
void rec##f() { \
	FlushRegs(); \
 \
	MOV32ItoR(EAX, (unsigned long)cpuRegs.code); \
	MOV32RtoM((unsigned long)&cpuRegs.code, EAX); \
	MOV32ItoR(EAX, (unsigned long)pc); \
	MOV32RtoM((unsigned long)&cpuRegs.pc, EAX); \
	MOV32ItoR(EAX, (unsigned long)int##f); \
	CALL32R  (EAX); \
	MOV32ItoR(EAX, 0); \
	MOV32RtoM((unsigned long)&cpuRegs.GPR.n.r0, EAX); \
	branch = 2; \
}

#define REC_BRANCH(f) \
void int##f(); \
void rec##f() { \
	FlushRegs(); \
 \
	MOV32ItoR(EAX, (unsigned long)cpuRegs.code); \
	MOV32RtoM((unsigned long)&cpuRegs.code, EAX); \
	MOV32ItoR(EAX, (unsigned long)pc); \
	MOV32RtoM((unsigned long)&cpuRegs.pc, EAX); \
	MOV32ItoR(EAX, (unsigned long)int##f); \
	CALL32R  (EAX); \
	MOV32ItoR(EAX, 0); \
	MOV32RtoM((unsigned long)&cpuRegs.GPR.n.r0, EAX); \
	branch = 2; \
	count++; \
}

static void recRecompile();

int recInit() {
	int i;

	recLUT = (unsigned long*) malloc(0x010000 * 4);

	recMem = (char*) malloc(0x800000); // 8mb
	recRAM = (char*) malloc(0x200000);
	recROM = (char*) malloc(0x080000);
	if (recRAM == NULL || recROM == NULL || recMem == NULL || recLUT == NULL) {
		SysMessage("Error allocating memory"); return -1;
	}

	for (i=0; i<0x80; i++) recLUT[i + 0x0000] = (unsigned long)&recRAM[(i & 0x1f) << 16];
	memcpy(recLUT + 0x8000, recLUT, 0x80 * 4);
	memcpy(recLUT + 0xa000, recLUT, 0x80 * 4);

	for (i=0; i<0x08; i++) recLUT[i + 0xbfc0] = (unsigned long)&recROM[i << 16];

	return 0;
}

void recReset() {
	memset(recRAM, 0, 0x200000);
	memset(recROM, 0, 0x080000);

	recPtr = recMem + 1024;
	branch = 0;
#ifdef CACHE_REGS
	memset(regs, 0, sizeof(regs));
	memset(regsmode, 0, sizeof(regsmode));
	regsm = 0;
	memset(x86regs, 0, sizeof(x86regs));
	x86regsc = 1;
	x86regsm = 0;
	x86skip  = 0;
#endif
}

void recShutdown() {
	if (recMem == NULL) return;
	free(recLUT);
	free(recMem);
	free(recRAM);
	free(recROM);
}

void recError() {
	SysReset();
	ClosePlugins();
	SysMessage("Unrecoverable error while running recompiler\n");
	SysRunGui();
}

#define execute() { \
	void (**recFunc)(); \
	char *p; \
 \
	p =	(char*)PC_REC(cpuRegs.pc); \
	if (p != NULL) recFunc = (void (**)()) (unsigned long)p; \
	else { recError(); return; } \
/*printf ("exec %x\n",cpuRegs.pc);*/ \
	if (*recFunc) { \
		(*recFunc)(); \
	} else { \
		recRecompile(); \
		(*recFunc)(); \
	} \
}

void DumpRegs() {
	int i, j;

	printf("%lx %lx\n", cpuRegs.pc, cpuIrq.cycle);
	for (i=0; i<4; i++) {
		for (j=0; j<8; j++)
			printf("%lx ", cpuRegs.GPR.r[j*i]);
		printf("\n");
	}
}

void recExecuteBios() {
	mode = 1;
	while (cpuRegs.pc != 0x80030000) {
		execute();
//		DumpRegs();
	}
}

void recExecute() {
/*	void (*recFunc)();
	unsigned long addr;

	mode = 1;
	recFunc = (void (*)()) ((unsigned long)recMem);
	recPtr = recMem;

// (recLUT[x >> 16] + (x & 0xffff))
	MOV32MtoR(EAX, (unsigned long)&cpuRegs.pc);
	MOV32RtoR(EDX, EAX);
	SHR32ItoR(EAX, 16);
	AND32ItoR(EDX, 0xffff);
	MOV32ItoR(ECX, (unsigned long)recLUT);
	SHL32ItoR(EAX, 2);
	ADD32RtoR(ECX, EAX);
	MOV32RmtoR(EAX, ECX);
	ADD32RtoR(EAX, EDX);
	MOV32RtoM((unsigned long)&addr, EAX);
	MOV32RmtoR(EAX, EAX);

	CMP32ItoR(EAX, 0);
	JNE8     (5 + 2);
	MOV32ItoR(EAX, (unsigned long)recRecompile);
	CALL32R  (EAX);
	MOV32MtoR(EAX, (unsigned long)&addr);
	MOV32RmtoR(EAX, EAX);
//	JMP32R   (EAX);
	CALL32R  (EAX);

	RET();
{
FILE *f;
f=fopen("log","w");
fwrite(recMem, 1, recPtr - recMem, f);
fclose(f);
}
	for (;;) recFunc();*/
	mode = 1;
	for (;;) execute();
}

void recClear(unsigned long Addr, unsigned long Size) {
	memset((void*)PC_REC(Addr), 0, Size * 4);
}

void recNULL() {
//	SysMessage("recUNK: %8.8x\n", cpuRegs.code);
}

/*********************************************************
* goes to opcodes tables...                              *
* Format:  table[something....]                          *
*********************************************************/

//REC_SYS(SPECIAL);
void recSPECIAL() {
	recSPC[_Funct_]();
}

void recREGIMM() {
	recREG[_Rt_]();
}

void recCOP0() {
	recCP0[_Rs_]();
}

//REC_SYS(COP2);
void recCOP2() {
	recCP2[_Funct_]();
}

void recBASIC() {
	recCP2BSC[_Rs_]();
}

//end of Tables opcodes...

/*********************************************************
* Arithmetic with immediate operand                      *
* Format:  OP rt, rs, immediate                          *
*********************************************************/
/*
REC_FUNC(ADDI);
REC_FUNC(ADDIU);
REC_FUNC(ANDI);
REC_FUNC(ORI);
REC_FUNC(XORI);
REC_FUNC(SLTI);
REC_FUNC(SLTIU);
#if 0*/
void recADDI()  {
// Rt = Rs + Im
	if (!_Rt_) return;

#ifdef CACHE_REGS
	CheckReg(_Rt_, RM_W);

//	if (_Rs_ != 0) {
		CheckReg(_Rs_, RM_R);

		MOV32RtoR(EAX, regs[_Rs_]);		// mov rt, rs
		ADD32ItoR(EAX, _Imm_);			// add rt, imm
		MOV32RtoR(regs[_Rt_], EAX);		// mov rt, rs
//		MOV32RtoR(regs[_Rt_], regs[_Rs_]);		// mov rt, rs
//		ADD32ItoR(regs[_Rt_], _Imm_);			// add rt, imm
/*	} else {
		MOV32ItoR(regs[_Rt_], _Imm_);			// mov rt, imm
	}*/
#else
	MOV32MtoR(EAX, (unsigned long)&cpuRegs.GPR.r[_Rs_]);
	ADD32ItoR(EAX, _Imm_);
	MOV32RtoM((unsigned long)&cpuRegs.GPR.r[_Rt_], EAX);
#endif
}

void recADDIU()  {
// Rt = Rs + Im
	if (!_Rt_) return;

#ifdef CACHE_REGS
	CheckReg(_Rt_, RM_W);

//	if (_Rs_ != 0) {
		CheckReg(_Rs_, RM_R);

		MOV32RtoR(EAX, regs[_Rs_]);		// mov rt, rs
		ADD32ItoR(EAX, _Imm_);			// add rt, imm
		MOV32RtoR(regs[_Rt_], EAX);		// mov rt, rs
//		MOV32RtoR(regs[_Rt_], regs[_Rs_]);		// mov rt, rs
//		ADD32ItoR(regs[_Rt_], _Imm_);			// add rt, imm
/*	} else {
		MOV32ItoR(regs[_Rt_], _Imm_);			// mov rt, imm
	}*/
#else
	MOV32MtoR(EAX, (unsigned long)&cpuRegs.GPR.r[_Rs_]);
	ADD32ItoR(EAX, _Imm_);
	MOV32RtoM((unsigned long)&cpuRegs.GPR.r[_Rt_], EAX);
#endif
}

void recSLTI() {
// Rt = Rs < Im (signed)
	if (!_Rt_) return;

#ifdef CACHE_REGS
    CheckReg(_Rt_, RM_W); CheckReg(_Rs_, RM_R);

/*    MOV32RtoR(regs[_Rt_], regs[_Rs_]);
    CMP32ItoR(regs[_Rt_], _Imm_);
    SETL8R   (regs[_Rt_]);
	MOV32RtoR(regs[_Rt_], EAX);
    AND32ItoR(regs[_Rt_], 0xff);*/

    MOV32RtoR(EAX, regs[_Rs_]);
    CMP32ItoR(EAX, _Imm_);
    SETL8R   (EAX);
    AND32ItoR(EAX, 0xff);
    MOV32RtoR(regs[_Rt_], EAX);
#else
	MOV32MtoR(EAX, (unsigned long)&cpuRegs.GPR.r[_Rs_]);
    CMP32ItoR(EAX, _Imm_);
    SETL8R   (EAX);
    AND32ItoR(EAX, 0xff);
	MOV32RtoM((unsigned long)&cpuRegs.GPR.r[_Rt_], EAX);
#endif
}

void recSLTIU() {
// Rt = Rs < Im (unsigned)
	if (!_Rt_) return;

#ifdef CACHE_REGS
    CheckReg(_Rt_, RM_W); CheckReg(_Rs_, RM_R);

/*    MOV32RtoR(regs[_Rt_], regs[_Rs_]);
    CMP32ItoR(regs[_Rt_], _Imm_);
	SBB32RtoR(regs[_Rt_], regs[_Rt_]);
	NEG32R   (regs[_Rt_]);
//    SETL8R   (regs[_Rt_]);
//    AND32ItoR(regs[_Rt_], 0xff);*/
    MOV32RtoR(EAX, regs[_Rs_]);
    CMP32ItoR(EAX, _Imm_);
	SBB32RtoR(EAX, EAX);
	NEG32R   (EAX);
	MOV32RtoR(regs[_Rt_], EAX);
#else
	MOV32MtoR(EAX, (unsigned long)&cpuRegs.GPR.r[_Rs_]);
    CMP32ItoR(EAX, _Imm_);
	SBB32RtoR(EAX, EAX);
	NEG32R   (EAX);
	MOV32RtoM((unsigned long)&cpuRegs.GPR.r[_Rt_], EAX);
#endif
}

void recANDI() {
// Rt = Rs And Im
	if (!_Rt_) return;

#ifdef CACHE_REGS
	CheckReg(_Rt_, RM_W);

//	if (_Rs_ != 0) {
		CheckReg(_Rs_, RM_R);

		MOV32RtoR(EAX, regs[_Rs_]);		// mov rt, rs
		AND32ItoR(EAX, _ImmU_);			// and rt, immu
		MOV32RtoR(regs[_Rt_], EAX);		// mov rt, rs
//		MOV32RtoR(regs[_Rt_], regs[_Rs_]);		// mov rt, rs
//		AND32ItoR(regs[_Rt_], _ImmU_);			// and rt, immu
/*	} else {
		MOV32ItoR(regs[_Rt_], 0);				// mov rt, 0
	}*/
#else
	MOV32MtoR(EAX, (unsigned long)&cpuRegs.GPR.r[_Rs_]);
	AND32ItoR(EAX, _ImmU_);
	MOV32RtoM((unsigned long)&cpuRegs.GPR.r[_Rt_], EAX);
#endif
}

void recORI() {
// Rt = Rs Or Im
	if (!_Rt_) return;

#ifdef CACHE_REGS
	CheckReg(_Rt_, RM_W);

//	if (_Rs_ != 0) {
		CheckReg(_Rs_, RM_R);

		MOV32RtoR(EAX, regs[_Rs_]);		// mov rt, rs
		OR32ItoR (EAX, _ImmU_);			// or  rt, immu
		MOV32RtoR(regs[_Rt_], EAX);		// mov rt, rs
//		MOV32RtoR(regs[_Rt_], regs[_Rs_]);		// mov rt, rs
//		OR32ItoR (regs[_Rt_], _ImmU_);			// or  rt, immu
/*	} else {
		MOV32ItoR(regs[_Rt_], _ImmU_);			// mov rt, immu
	}*/
#else
	MOV32MtoR(EAX, (unsigned long)&cpuRegs.GPR.r[_Rs_]);
	OR32ItoR (EAX, _ImmU_);
	MOV32RtoM((unsigned long)&cpuRegs.GPR.r[_Rt_], EAX);
#endif
}


void recXORI() {
// Rt = Rs Xor Im
	if (!_Rt_) return;

#ifdef CACHE_REGS
    CheckReg(_Rt_, RM_W);

//	if (_Rs_ != 0) {
		CheckReg(_Rs_, RM_R);

		MOV32RtoR(EAX, regs[_Rs_]);		// mov rt, rs
		XOR32ItoR(EAX, _ImmU_);			// xor rt, immu
		MOV32RtoR(regs[_Rt_], EAX);		// mov rt, rs
//		MOV32RtoR(regs[_Rt_], regs[_Rs_]);		// mov rt, rs
//		XOR32ItoR(regs[_Rt_], _ImmU_);			// xor rt, immu
/*	} else {
		MOV32ItoR(regs[_Rt_], _ImmU_ ^ 0);		// mov rt, immu ^ 0
	}*/
#else
	MOV32MtoR(EAX, (unsigned long)&cpuRegs.GPR.r[_Rs_]);
	XOR32ItoR(EAX, _ImmU_);
	MOV32RtoM((unsigned long)&cpuRegs.GPR.r[_Rt_], EAX);
#endif
}
//#endif
//end of * Arithmetic with immediate operand  

/*********************************************************
* Load higher 16 bits of the first word in GPR with imm  *
* Format:  OP rt, immediate                              *
*********************************************************/
/*REC_FUNC(LUI);
#if 0*/
void recLUI()  {
// Rt = Imm << 16
	if (!_Rt_) return;

#ifdef CACHE_REGS
	CheckReg(_Rt_, RM_W);

	MOV32ItoR(regs[_Rt_], cpuRegs.code << 16); // mov rt, inst << 16
#else
	MOV32ItoM((unsigned long)&cpuRegs.GPR.r[_Rt_], cpuRegs.code << 16);
#endif
}
//#endif
//End of Load Higher .....


/*********************************************************
* Register arithmetic                                    *
* Format:  OP rd, rs, rt                                 *
*********************************************************/
/*
REC_FUNC(ADD);
REC_FUNC(ADDU);
REC_FUNC(SUB);
REC_FUNC(SUBU);
REC_FUNC(AND);
REC_FUNC(OR);
REC_FUNC(XOR);
REC_FUNC(NOR);
REC_FUNC(SLT);
REC_FUNC(SLTU);

#if 0*/
void recADD() {
// Rd = Rs + Rt
	if (!_Rd_) return;

#ifdef CACHE_REGS
	CheckReg(_Rd_, RM_W);

//	if ((_Rs_ != 0) && (_Rt_ != 0)) {
		CheckReg(_Rt_, RM_R); CheckReg(_Rs_, RM_R);

		MOV32RtoR(EAX, regs[_Rs_]);				// mov eax, rs
		ADD32RtoR(EAX, regs[_Rt_]);				// add eax, rt
    	MOV32RtoR(regs[_Rd_], EAX);				// mov rd, eax
/*	} else if (_Rs_ == 0) {
		CheckReg(_Rt_, RM_R);

    	MOV32RtoR(regs[_Rd_], regs[_Rt_]);		// mov rd, rt
	} else if (_Rt_ == 0) {
		CheckReg(_Rs_, RM_R);

    	MOV32RtoR(regs[_Rd_], regs[_Rs_]);		// mov rd, rs
	} else {
    	MOV32ItoR(regs[_Rd_], 0);				// mov rd, 0
	}*/
#else
	MOV32MtoR(EAX, (unsigned long)&cpuRegs.GPR.r[_Rs_]);
	ADD32MtoR(EAX, (unsigned long)&cpuRegs.GPR.r[_Rt_]);
	MOV32RtoM((unsigned long)&cpuRegs.GPR.r[_Rd_], EAX);
#endif
}

void recADDU() {
// Rd = Rs + Rt 
	if (!_Rd_) return;

#ifdef CACHE_REGS
//	FlushRegs();

	CheckReg(_Rd_, RM_W);

//	if ((_Rs_ != 0) && (_Rt_ != 0)) {
		CheckReg(_Rt_, RM_R); CheckReg(_Rs_, RM_R);

		MOV32RtoR(EAX, regs[_Rs_]);				// mov eax, rs
		ADD32RtoR(EAX, regs[_Rt_]);				// add eax, rt
    	MOV32RtoR(regs[_Rd_], EAX);				// mov rd, eax
/*	} else if (_Rs_ == 0) {
		CheckReg(_Rt_, RM_R);

    	MOV32RtoR(regs[_Rd_], regs[_Rt_]);		// mov rd, rt
	} else if (_Rt_ == 0) {
		CheckReg(_Rs_, RM_R);

    	MOV32RtoR(regs[_Rd_], regs[_Rs_]);		// mov rd, rs
	} else {
    	MOV32ItoR(regs[_Rd_], 0);				// mov rd, 0
	}*/
#else
	MOV32MtoR(EAX, (unsigned long)&cpuRegs.GPR.r[_Rs_]);
	ADD32MtoR(EAX, (unsigned long)&cpuRegs.GPR.r[_Rt_]);
	MOV32RtoM((unsigned long)&cpuRegs.GPR.r[_Rd_], EAX);
#endif
}

void recSUB() {
// Rd = Rs - Rt
	if (!_Rd_) return;

#ifdef CACHE_REGS
	CheckReg(_Rd_, RM_W);

//	if ((_Rs_ != 0) && (_Rt_ != 0)) {
		CheckReg(_Rt_, RM_R); CheckReg(_Rs_, RM_R);

		MOV32RtoR(EAX, regs[_Rs_]);				// mov eax, rs
		SUB32RtoR(EAX, regs[_Rt_]);				// sub eax, rt
    	MOV32RtoR(regs[_Rd_], EAX);				// mov rd, eax
/*	} else if (_Rs_ == 0) {
		CheckReg(_Rt_, RM_R);

		MOV32ItoR(EAX, 0);						// mov eax, 0
		SUB32RtoR(EAX, regs[_Rt_]);				// sub eax, rt
    	MOV32RtoR(regs[_Rd_], EAX);				// mov rd, eax
	} else if (_Rt_ == 0) {
		CheckReg(_Rs_, RM_R);

    	MOV32RtoR(regs[_Rd_], regs[_Rs_]);		// mov rd, rs
	} else {
    	MOV32ItoR(regs[_Rd_], 0);				// mov rd, 0
	}*/
#else
	MOV32MtoR(EAX, (unsigned long)&cpuRegs.GPR.r[_Rs_]);
	SUB32MtoR(EAX, (unsigned long)&cpuRegs.GPR.r[_Rt_]);
	MOV32RtoM((unsigned long)&cpuRegs.GPR.r[_Rd_], EAX);
#endif
}   

void recSUBU() {
// Rd = Rs - Rt
	if (!_Rd_) return;

#ifdef CACHE_REGS
	CheckReg(_Rd_, RM_W);

//	if ((_Rs_ != 0) && (_Rt_ != 0)) {
		CheckReg(_Rt_, RM_R); CheckReg(_Rs_, RM_R);

		MOV32RtoR(EAX, regs[_Rs_]);				// mov eax, rs
		SUB32RtoR(EAX, regs[_Rt_]);				// sub eax, rt
    	MOV32RtoR(regs[_Rd_], EAX);				// mov rd, eax
/*	} else if (_Rs_ == 0) {
		CheckReg(_Rt_, RM_R);

		MOV32ItoR(EAX, 0);						// mov eax, 0
		SUB32RtoR(EAX, regs[_Rt_]);				// sub eax, rt
    	MOV32RtoR(regs[_Rd_], EAX);				// mov rd, eax
	} else if (_Rt_ == 0) {
		CheckReg(_Rs_, RM_R);

    	MOV32RtoR(regs[_Rd_], regs[_Rs_]);		// mov rd, rs
	} else {
    	MOV32ItoR(regs[_Rd_], 0);				// mov rd, 0
	}*/
#else
	MOV32MtoR(EAX, (unsigned long)&cpuRegs.GPR.r[_Rs_]);
	SUB32MtoR(EAX, (unsigned long)&cpuRegs.GPR.r[_Rt_]);
	MOV32RtoM((unsigned long)&cpuRegs.GPR.r[_Rd_], EAX);
#endif
}   

void recAND() {
// Rd = Rs And Rt
	if (!_Rd_) return;

#ifdef CACHE_REGS
	CheckReg(_Rd_, RM_W);

//	if ((_Rs_ != 0) && (_Rt_ != 0)) {
		CheckReg(_Rt_, RM_R); CheckReg(_Rs_, RM_R);

		MOV32RtoR(EAX, regs[_Rs_]);				// mov eax, rs
		AND32RtoR(EAX, regs[_Rt_]);				// and eax, rt
    	MOV32RtoR(regs[_Rd_], EAX);				// mov rd, eax
/*	} else {
    	MOV32ItoR(regs[_Rd_], 0);				// mov rd, 0
	}*/
#else
	MOV32MtoR(EAX, (unsigned long)&cpuRegs.GPR.r[_Rs_]);
	AND32MtoR(EAX, (unsigned long)&cpuRegs.GPR.r[_Rt_]);
	MOV32RtoM((unsigned long)&cpuRegs.GPR.r[_Rd_], EAX);
#endif
}   

void recOR() {
// Rd = Rs Or Rt
	if (!_Rd_) return;

#ifdef CACHE_REGS
	CheckReg(_Rd_, RM_W);

//	if ((_Rs_ != 0) && (_Rt_ != 0)) {
		CheckReg(_Rt_, RM_R); CheckReg(_Rs_, RM_R);

		MOV32RtoR(EAX, regs[_Rs_]);				// mov eax, rs
		OR32RtoR (EAX, regs[_Rt_]);				// or  eax, rt
    	MOV32RtoR(regs[_Rd_], EAX);				// mov rd, eax
/*	} else if (_Rs_ == 0) {
		CheckReg(_Rt_, RM_R);

    	MOV32RtoR(regs[_Rd_], regs[_Rt_]);		// mov rd, rt
	} else if (_Rt_ == 0) {
		CheckReg(_Rs_, RM_R);

    	MOV32RtoR(regs[_Rd_], regs[_Rs_]);		// mov rd, rs
	} else {
    	MOV32ItoR(regs[_Rd_], 0);				// mov rd, 0
	}*/
#else
	MOV32MtoR(EAX, (unsigned long)&cpuRegs.GPR.r[_Rs_]);
	OR32MtoR (EAX, (unsigned long)&cpuRegs.GPR.r[_Rt_]);
	MOV32RtoM((unsigned long)&cpuRegs.GPR.r[_Rd_], EAX);
#endif
}   

void recXOR() {
// Rd = Rs Xor Rt
	if (!_Rd_) return;

#ifdef CACHE_REGS
	CheckReg(_Rd_, RM_W);

//	if ((_Rs_ != 0) && (_Rt_ != 0)) {
		CheckReg(_Rt_, RM_R); CheckReg(_Rs_, RM_R);

		MOV32RtoR(EAX, regs[_Rs_]);				// mov eax, rs
		XOR32RtoR(EAX, regs[_Rt_]);				// xor eax, rt
    	MOV32RtoR(regs[_Rd_], EAX);				// mov rd, eax
/*	} else if (_Rs_ == 0) {
		CheckReg(_Rt_, RM_R);

		MOV32ItoR(EAX, 0);						// mov eax, 0
		XOR32RtoR(EAX, regs[_Rt_]);				// xor eax, rt
    	MOV32RtoR(regs[_Rd_], regs[_Rt_]);		// mov rd, rt
	} else if (_Rt_ == 0) {
		CheckReg(_Rs_, RM_R);

		MOV32RtoR(EAX, regs[_Rs_]);				// mov eax, rs
		XOR32ItoR(EAX, 0);						// xor eax, 0
    	MOV32RtoR(regs[_Rd_], regs[_Rs_]);		// mov rd, rs
	} else {
    	MOV32ItoR(regs[_Rd_], 0);				// mov rd, 0
	}*/
#else
	MOV32MtoR(EAX, (unsigned long)&cpuRegs.GPR.r[_Rs_]);
	XOR32MtoR(EAX, (unsigned long)&cpuRegs.GPR.r[_Rt_]);
	MOV32RtoM((unsigned long)&cpuRegs.GPR.r[_Rd_], EAX);
#endif
}

void recNOR() {
// Rd = Rs Nor Rt
	if (!_Rd_) return;

#ifdef CACHE_REGS
	CheckReg(_Rd_, RM_W);

//	if ((_Rs_ != 0) && (_Rt_ != 0)) {
		CheckReg(_Rt_, RM_R); CheckReg(_Rs_, RM_R);

		MOV32RtoR(EAX, regs[_Rs_]);				// mov eax, rs
		OR32RtoR (EAX, regs[_Rt_]);				// or  eax, rt
		NOT32R   (EAX);							// not eax
    	MOV32RtoR(regs[_Rd_], EAX);				// mov rd, eax
/*	} else if (_Rs_ == 0) {
		CheckReg(_Rt_, RM_R);

    	MOV32RtoR(regs[_Rd_], regs[_Rt_]);		// mov rd, rt
		NOT32R   (regs[_Rd_]);					// not rd
	} else if (_Rt_ == 0) {
		CheckReg(_Rs_, RM_R);

    	MOV32RtoR(regs[_Rd_], regs[_Rs_]);		// mov rd, rs
		NOT32R   (regs[_Rd_]);					// not rd
	} else {
    	MOV32ItoR(regs[_Rd_], ~0);				// mov rd, ~0
	}*/
#else
	MOV32MtoR(EAX, (unsigned long)&cpuRegs.GPR.r[_Rs_]);
	OR32MtoR (EAX, (unsigned long)&cpuRegs.GPR.r[_Rt_]);
	NOT32R   (EAX);
	MOV32RtoM((unsigned long)&cpuRegs.GPR.r[_Rd_], EAX);
#endif
}

void recSLT() {
// Rd = Rs < Rt (signed)
	if (!_Rd_) return;

#ifdef CACHE_REGS
    CheckReg(_Rd_, RM_W);

//	if ((_Rs_ != 0) && (_Rt_ != 0)) {
		CheckReg(_Rt_, RM_R); CheckReg(_Rs_, RM_R);

		MOV32RtoR(EAX, regs[_Rs_]);
    	CMP32RtoR(EAX, regs[_Rt_]);
    	SETL8R   (EAX);
    	AND32ItoR(EAX, 0xff);
		MOV32RtoR(regs[_Rd_], EAX);
/*	} else if (_Rs_ == 0) {
		CheckReg(_Rt_, RM_R);

		MOV32ItoR(EAX, 0);
    	CMP32RtoR(EAX, regs[_Rt_]);
    	SETL8R   (EAX);
    	AND32ItoR(EAX, 0xff);
		MOV32RtoR(regs[_Rd_], EAX);
	} else if (_Rt_ == 0) {
		CheckReg(_Rs_, RM_R);

    	CMP32ItoR(regs[_Rs_], 0);
    	SETL8R   (EAX);
    	AND32ItoR(EAX, 0xff);
		MOV32RtoR(regs[_Rd_], EAX);
	} else {
    	MOV32ItoR(regs[_Rd_], 0);				// mov rd, 0
	}*/
#else
	MOV32MtoR(EAX, (unsigned long)&cpuRegs.GPR.r[_Rs_]);
    CMP32MtoR(EAX, (unsigned long)&cpuRegs.GPR.r[_Rt_]);
    SETL8R   (EAX);
    AND32ItoR(EAX, 0xff);
	MOV32RtoM((unsigned long)&cpuRegs.GPR.r[_Rd_], EAX);
#endif
}  

void recSLTU() { 
// Rd = Rs < Rt (unsigned)
	if (!_Rd_) return;

#ifdef CACHE_REGS
    CheckReg(_Rd_, RM_W);

//	if ((_Rs_ != 0) && (_Rt_ != 0)) {
		CheckReg(_Rt_, RM_R); CheckReg(_Rs_, RM_R);

		MOV32RtoR(EAX, regs[_Rs_]);
    	CMP32RtoR(EAX, regs[_Rt_]);
		SBB32RtoR(EAX, EAX);
		NEG32R   (EAX);
//    	SETL8R   (EAX);
//    	AND32ItoR(EAX, 0xff);
		MOV32RtoR(regs[_Rd_], EAX);
/*	} else if (_Rs_ == 0) {
		CheckReg(_Rt_, RM_R);

		MOV32ItoR(EAX, 0);
    	CMP32RtoR(EAX, regs[_Rt_]);
		SBB32RtoR(EAX, EAX);
		NEG32R   (EAX);
//    	SETL8R   (EAX);
//    	AND32ItoR(EAX, 0xff);
		MOV32RtoR(regs[_Rd_], EAX);
	} else if (_Rt_ == 0) {
		CheckReg(_Rs_, RM_R);

    	CMP32ItoR(regs[_Rs_], 0);
		SBB32RtoR(EAX, EAX);
		NEG32R   (EAX);
//    	SETL8R   (EAX);
//    	AND32ItoR(EAX, 0xff);
		MOV32RtoR(regs[_Rd_], EAX);
	} else {
    	MOV32ItoR(regs[_Rd_], 0);				// mov rd, 0
	}*/
#else
	MOV32MtoR(EAX, (unsigned long)&cpuRegs.GPR.r[_Rs_]);
    CMP32MtoR(EAX, (unsigned long)&cpuRegs.GPR.r[_Rt_]);
	SBB32RtoR(EAX, EAX);
	NEG32R   (EAX);
	MOV32RtoM((unsigned long)&cpuRegs.GPR.r[_Rd_], EAX);
#endif
}
//#endif
//End of * Register arithmetic

/*********************************************************
* Register mult/div & Register trap logic                *
* Format:  OP rs, rt                                     *
*********************************************************/

/*REC_FUNC(MULT);
REC_FUNC(MULTU);
REC_FUNC(DIV);
REC_FUNC(DIVU);
#if 0*/
void recMULT() {
// Lo/Hi = Rs * Rt (signed)
#ifdef CACHE_REGS
	FlushRegs();

	x86skip = EDX;
	x86regsc = ECX;
    CheckReg(_Rt_, RM_R); CheckReg(_Rs_, RM_R);
	x86skip = 0;

	MOV32RtoR(EAX, regs[_Rs_]);
	IMUL32R  (regs[_Rt_]);
	MOV32RtoM((unsigned long)&cpuRegs.GPR.n.lo, EAX);
	MOV32RtoM((unsigned long)&cpuRegs.GPR.n.hi, EDX);
#else
	MOV32MtoR(EAX, (unsigned long)&cpuRegs.GPR.r[_Rs_]);
	IMUL32M  ((unsigned long)&cpuRegs.GPR.r[_Rt_]);
	MOV32RtoM((unsigned long)&cpuRegs.GPR.n.lo, EAX);
	MOV32RtoM((unsigned long)&cpuRegs.GPR.n.hi, EDX);
#endif
}

void recMULTU() {
// Lo/Hi = Rs * Rt (unsigned)
#ifdef CACHE_REGS
	FlushRegs();

	x86skip = EDX;
	x86regsc = ECX;
    CheckReg(_Rt_, RM_R); CheckReg(_Rs_, RM_R);
	x86skip = 0;

	MOV32RtoR(EAX, regs[_Rs_]);
	MUL32R   (regs[_Rt_]);
	MOV32RtoM((unsigned long)&cpuRegs.GPR.n.lo, EAX);
	MOV32RtoM((unsigned long)&cpuRegs.GPR.n.hi, EDX);
#else
	MOV32MtoR(EAX, (unsigned long)&cpuRegs.GPR.r[_Rs_]);
	MUL32M   ((unsigned long)&cpuRegs.GPR.r[_Rt_]);
	MOV32RtoM((unsigned long)&cpuRegs.GPR.n.lo, EAX);
	MOV32RtoM((unsigned long)&cpuRegs.GPR.n.hi, EDX);
#endif
}

void recDIV() {
// Lo/Hi = Rs / Rt (signed)
#ifdef CACHE_REGS
	FlushRegs();
//	ClearX86Reg(EDX);

	x86skip = EDX;
	x86regsc = ECX;
    CheckReg(_Rt_, RM_R); CheckReg(_Rs_, RM_R);
	x86skip = 0;

	MOV32RtoR(EAX, regs[_Rs_]);
	CDQ();
	IDIV32R  (regs[_Rt_]);
	MOV32RtoM((unsigned long)&cpuRegs.GPR.n.lo, EAX);
	MOV32RtoM((unsigned long)&cpuRegs.GPR.n.hi, EDX);
#else
	MOV32MtoR(ECX, (unsigned long)&cpuRegs.GPR.r[_Rt_]);
	CMP32ItoR(ECX, 0);
	JE8      (6 + 1 + 2 + 6 + 6);
	MOV32MtoR(EAX, (unsigned long)&cpuRegs.GPR.r[_Rs_]);
	CDQ();
	IDIV32R  (ECX);
	MOV32RtoM((unsigned long)&cpuRegs.GPR.n.lo, EAX);
	MOV32RtoM((unsigned long)&cpuRegs.GPR.n.hi, EDX);
#endif
}

void recDIVU() {
// Lo/Hi = Rs / Rt (unsigned)
#ifdef CACHE_REGS
	FlushRegs();
//	ClearX86Reg(EDX);

	x86skip = EDX;
	x86regsc = ECX;
    CheckReg(_Rt_, RM_R); CheckReg(_Rs_, RM_R);
	x86skip = 0;

	MOV32RtoR(EAX, regs[_Rs_]);
	XOR32RtoR(EDX, EDX);
	DIV32R   (regs[_Rt_]);
	MOV32RtoM((unsigned long)&cpuRegs.GPR.n.lo, EAX);
	MOV32RtoM((unsigned long)&cpuRegs.GPR.n.hi, EDX);
#else
	MOV32MtoR(ECX, (unsigned long)&cpuRegs.GPR.r[_Rt_]);
	CMP32ItoR(ECX, 0);
	JE8      (6 + 2 + 2 + 6 + 6);
	MOV32MtoR(EAX, (unsigned long)&cpuRegs.GPR.r[_Rs_]);
	XOR32RtoR(EDX, EDX);
	DIV32R   (ECX);
	MOV32RtoM((unsigned long)&cpuRegs.GPR.n.lo, EAX);
	MOV32RtoM((unsigned long)&cpuRegs.GPR.n.hi, EDX);
#endif
}
//#endif
//End of * Register mult/div & Register trap logic  

/*REC_FUNC(LB);
REC_FUNC(LBU);
REC_FUNC(LH);
REC_FUNC(LHU);
REC_FUNC(LW);

REC_FUNC(SB);
REC_FUNC(SH);
REC_FUNC(SW);*/

REC_FUNC(LWL);
REC_FUNC(LWR);
REC_FUNC(SWL);
REC_FUNC(SWR);
//#if 0
void recLB() {
// Rt = mem[Rs + Im] (signed)

#ifdef CACHE_REGS
//	if (_Rs_ != 0) {
	    CheckReg(_Rs_, RM_R);

		MOV32RtoR(EAX, regs[_Rs_]);
		ADD32ItoR(EAX, _Imm_);
		PUSH32R  (EAX);
/*	} else {
		PUSH32I  (_Imm_);
	}*/

	FlushRegs();
	MOV32ItoR(EAX, (unsigned long)memRead8);
	CALL32R  (EAX);
	ADD32ItoR(ESP, 4);
	if (_Rt_) {
		CheckReg(_Rt_, RM_W);
		CBW(); CWD();
		MOV32RtoR(regs[_Rt_], EAX);
	}
#else
	MOV32MtoR(EAX, (unsigned long)&cpuRegs.GPR.r[_Rs_]);
	ADD32ItoR(EAX, _Imm_);
	PUSH32R  (EAX);
	MOV32ItoR(EAX, (unsigned long)memRead8);
	CALL32R  (EAX);
	if (_Rt_) {
		MOVSX32R8toR(EAX, EAX);
//		CBW(); CWD();
		MOV32RtoM((unsigned long)&cpuRegs.GPR.r[_Rt_], EAX);
	}
	ADD32ItoR(ESP, 4);
#endif
}

void recLBU() {
// Rt = mem[Rs + Im] (unsigned)

#ifdef CACHE_REGS
//	if (_Rs_ != 0) {
	    CheckReg(_Rs_, RM_R);

		MOV32RtoR(EAX, regs[_Rs_]);
		ADD32ItoR(EAX, _Imm_);
		PUSH32R  (EAX);
/*	} else {
		PUSH32I  (_Imm_);
	}*/
	FlushRegs();
	MOV32ItoR(EAX, (unsigned long)memRead8);
	CALL32R  (EAX);
	ADD32ItoR(ESP, 4);
	if (_Rt_) {
		CheckReg(_Rt_, RM_W);
		MOV32RtoR(regs[_Rt_], EAX);
		AND32ItoR(regs[_Rt_], 0xff);
	}
#else
	MOV32MtoR(EAX, (unsigned long)&cpuRegs.GPR.r[_Rs_]);
	ADD32ItoR(EAX, _Imm_);
	PUSH32R  (EAX);
	MOV32ItoR(EAX, (unsigned long)memRead8);
	CALL32R  (EAX);
	if (_Rt_) {
//		AND32ItoR(EAX, 0xff);
		MOVZX32R8toR(EAX, EAX);
		MOV32RtoM((unsigned long)&cpuRegs.GPR.r[_Rt_], EAX);
	}
	ADD32ItoR(ESP, 4);
#endif
}

void recLH() {
// Rt = mem[Rs + Im] (signed)

#ifdef CACHE_REGS
//	if (_Rs_ != 0) {
	    CheckReg(_Rs_, RM_R);

		MOV32RtoR(EAX, regs[_Rs_]);
		ADD32ItoR(EAX, _Imm_);
		PUSH32R  (EAX);
/*	} else {
		PUSH32I  (_Imm_);
	}*/
	FlushRegs();
	MOV32ItoR(EAX, (unsigned long)memRead16);
	CALL32R  (EAX);
	ADD32ItoR(ESP, 4);
	if (_Rt_) {
		CheckReg(_Rt_, RM_W);
		CWD();
		MOV32RtoR(regs[_Rt_], EAX);
	}
#else
	MOV32MtoR(EAX, (unsigned long)&cpuRegs.GPR.r[_Rs_]);
	ADD32ItoR(EAX, _Imm_);
	PUSH32R  (EAX);
	MOV32ItoR(EAX, (unsigned long)memRead16);
	CALL32R  (EAX);
	if (_Rt_) {
//		CWD();
		MOVSX32R16toR(EAX, EAX);
		MOV32RtoM((unsigned long)&cpuRegs.GPR.r[_Rt_], EAX);
	}
	ADD32ItoR(ESP, 4);
#endif
}

void recLHU() {
// Rt = mem[Rs + Im] (unsigned)

#ifdef CACHE_REGS
//	if (_Rs_ != 0) {
	    CheckReg(_Rs_, RM_R);

		MOV32RtoR(EAX, regs[_Rs_]);
		ADD32ItoR(EAX, _Imm_);
		PUSH32R  (EAX);
/*	} else {
		PUSH32I  (_Imm_);
	}*/
	FlushRegs();
	MOV32ItoR(EAX, (unsigned long)memRead16);
	CALL32R  (EAX);
	ADD32ItoR(ESP, 4);
	if (_Rt_) {
		CheckReg(_Rt_, RM_W);
		MOV32RtoR(regs[_Rt_], EAX);
		AND32ItoR(regs[_Rt_], 0xffff);
	}
#else
	MOV32MtoR(EAX, (unsigned long)&cpuRegs.GPR.r[_Rs_]);
	ADD32ItoR(EAX, _Imm_);
	PUSH32R  (EAX);
	MOV32ItoR(EAX, (unsigned long)memRead16);
	CALL32R  (EAX);
	if (_Rt_) {
//		AND32ItoR(EAX, 0xffff);
		MOVZX32R16toR(EAX, EAX);
		MOV32RtoM((unsigned long)&cpuRegs.GPR.r[_Rt_], EAX);
	}
	ADD32ItoR(ESP, 4);
#endif
}

void recLW() {
// Rt = mem[Rs + Im] (unsigned)

#ifdef CACHE_REGS
//	if (_Rs_ != 0) {
	    CheckReg(_Rs_, RM_R);

		MOV32RtoR(EAX, regs[_Rs_]);
		ADD32ItoR(EAX, _Imm_);
		PUSH32R  (EAX);
/*	} else {
		PUSH32I  (_Imm_);
	}*/
	FlushRegs();
	MOV32ItoR(EAX, (unsigned long)memRead32);
	CALL32R  (EAX);
	ADD32ItoR(ESP, 4);
	if (_Rt_) {
		CheckReg(_Rt_, RM_W);
		MOV32RtoR(regs[_Rt_], EAX);
	}
#else
	MOV32MtoR(EAX, (unsigned long)&cpuRegs.GPR.r[_Rs_]);
	ADD32ItoR(EAX, _Imm_);
	PUSH32R  (EAX);
	MOV32ItoR(EAX, (unsigned long)memRead32);
	CALL32R  (EAX);
	if (_Rt_) {
		MOV32RtoM((unsigned long)&cpuRegs.GPR.r[_Rt_], EAX);
	}
	ADD32ItoR(ESP, 4);
#endif
}

/*
void recLWL() {
	printf ("LWL\n"); exit(0);
}

void recLWR() {
	printf ("LWR\n"); exit(0);
}*/

void recSB() {
// mem[Rt] = Rs + Im

#ifdef CACHE_REGS
//	if (_Rt_ != 0) {
    	CheckReg(_Rt_, RM_R);

		PUSH32R  (regs[_Rt_]);
/*	} else {
		PUSH32I  (0);
	}
	if (_Rs_ != 0) {*/
    	CheckReg(_Rs_, RM_R);

		MOV32RtoR(EAX, regs[_Rs_]);
//		if (_Imm_ != 0) {
			ADD32ItoR(EAX, _Imm_);
//		}
		PUSH32R  (EAX);
/*	} else {
		PUSH32I  (_Imm_);
	}*/
	FlushRegs();
	MOV32ItoR(EAX, (unsigned long)memWrite8);
	CALL32R  (EAX);
	ADD32ItoR(ESP, 8);
#else
	PUSH32M  ((unsigned long)&cpuRegs.GPR.r[_Rt_]);
	MOV32MtoR(EAX, (unsigned long)&cpuRegs.GPR.r[_Rs_]);
	ADD32ItoR(EAX, _Imm_);
	PUSH32R  (EAX);
	MOV32ItoR(EAX, (unsigned long)memWrite8);
	CALL32R  (EAX);
	ADD32ItoR(ESP, 8);
#endif
}

void recSH() {
// mem[Rt] = Rs + Im

#ifdef CACHE_REGS
//	if (_Rt_ != 0) {
    	CheckReg(_Rt_, RM_R);

		PUSH32R  (regs[_Rt_]);
/*	} else {
		PUSH32I  (0);
	}
	if (_Rs_ != 0) {*/
    	CheckReg(_Rs_, RM_R);

		MOV32RtoR(EAX, regs[_Rs_]);
//		if (_Imm_ != 0) {
			ADD32ItoR(EAX, _Imm_);
//		}
		PUSH32R  (EAX);
/*	} else {
		PUSH32I  (_Imm_);
	}*/
	FlushRegs();
	MOV32ItoR(EAX, (unsigned long)memWrite16);
	CALL32R  (EAX);
	ADD32ItoR(ESP, 8);
#else
	PUSH32M  ((unsigned long)&cpuRegs.GPR.r[_Rt_]);
	MOV32MtoR(EAX, (unsigned long)&cpuRegs.GPR.r[_Rs_]);
	ADD32ItoR(EAX, _Imm_);
	PUSH32R  (EAX);
	MOV32ItoR(EAX, (unsigned long)memWrite16);
	CALL32R  (EAX);
	ADD32ItoR(ESP, 8);
#endif
}

void recSW() {
// mem[Rt] = Rs + Im

#ifdef CACHE_REGS
/*    	CheckReg(_Rt_, RM_R);
    	CheckReg(_Rs_, RM_R);
	PUSHA32();

		PUSH32R  (regs[_Rt_]);


		MOV32RtoR(EAX, regs[_Rs_]);
		if (_Imm_ != 0)
			ADD32ItoR(EAX, _Imm_);
		PUSH32R  (EAX);*/
//	if (_Rt_ != 0) {
    	CheckReg(_Rt_, RM_R);

		PUSH32R  (regs[_Rt_]);
/*	} else {
		PUSH32I  (0);
	}
	if (_Rs_ != 0) {*/
    	CheckReg(_Rs_, RM_R);

		MOV32RtoR(EAX, regs[_Rs_]);
//		if (_Imm_ != 0)
			ADD32ItoR(EAX, _Imm_);
		PUSH32R  (EAX);
/*	} else {
		PUSH32I  (_Imm_);
	}*/
	FlushRegs();
	MOV32ItoR(EAX, (unsigned long)memWrite32);
	CALL32R  (EAX);
	ADD32ItoR(ESP, 8);
//	POPA32();
#else
	PUSH32M  ((unsigned long)&cpuRegs.GPR.r[_Rt_]);
	MOV32MtoR(EAX, (unsigned long)&cpuRegs.GPR.r[_Rs_]);
	ADD32ItoR(EAX, _Imm_);
	PUSH32R  (EAX);
	MOV32ItoR(EAX, (unsigned long)memWrite32);
	CALL32R  (EAX);
	ADD32ItoR(ESP, 8);
#endif
}
//#endif
/*
void recSWL() {
	printf ("SWL\n"); exit(0);
}

void recSWR() {
	printf ("SWR\n"); exit(0);
}*/

/*REC_FUNC(SLL);
REC_FUNC(SRL);
REC_FUNC(SRA);
#if 0*/
void recSLL() {
// Rd = Rt << Sa
	if (!_Rd_) return;

#ifdef CACHE_REGS
    CheckReg(_Rd_, RM_W); CheckReg(_Rt_, RM_R);

	MOV32RtoR(EAX, regs[_Rt_]);
	SHL32ItoR(EAX, _Sa_);
	MOV32RtoR(regs[_Rd_], EAX);
//	MOV32RtoR(regs[_Rd_], regs[_Rt_]);
//	SHL32ItoR(regs[_Rd_], _Sa_);
#else
	MOV32MtoR(EAX, (unsigned long)&cpuRegs.GPR.r[_Rt_]);
	SHL32ItoR(EAX, _Sa_);
	MOV32RtoM((unsigned long)&cpuRegs.GPR.r[_Rd_], EAX);
#endif
}

void recSRL() {
// Rd = Rt >> Sa
	if (!_Rd_) return;

#ifdef CACHE_REGS
    CheckReg(_Rd_, RM_W); CheckReg(_Rt_, RM_R);

	MOV32RtoR(EAX, regs[_Rt_]);
	SHR32ItoR(EAX, _Sa_);
	MOV32RtoR(regs[_Rd_], EAX);
//	MOV32RtoR(regs[_Rd_], regs[_Rt_]);
//	SHR32ItoR(regs[_Rd_], _Sa_);
#else
	MOV32MtoR(EAX, (unsigned long)&cpuRegs.GPR.r[_Rt_]);
	SHR32ItoR(EAX, _Sa_);
	MOV32RtoM((unsigned long)&cpuRegs.GPR.r[_Rd_], EAX);
#endif
}

void recSRA() {
// Rd = Rt >> Sa
	if (!_Rd_) return;

#ifdef CACHE_REGS
    CheckReg(_Rd_, RM_W); CheckReg(_Rt_, RM_R);

	MOV32RtoR(EAX, regs[_Rt_]);
	SAR32ItoR(EAX, _Sa_);
	MOV32RtoR(regs[_Rd_], EAX);
//	MOV32RtoR(regs[_Rd_], regs[_Rt_]);
//	SAR32ItoR(regs[_Rd_], _Sa_);
#else
	MOV32MtoR(EAX, (unsigned long)&cpuRegs.GPR.r[_Rt_]);
	SAR32ItoR(EAX, _Sa_);
	MOV32RtoM((unsigned long)&cpuRegs.GPR.r[_Rd_], EAX);
#endif
}
//#endif

/*REC_FUNC(SLLV);
REC_FUNC(SRLV);
REC_FUNC(SRAV);
#if 0*/
void recSLLV() {
// Rd = Rt << Rs
	if (!_Rd_) return;

#ifdef CACHE_REGS
	FlushRegs();

	x86skip = ECX;
	x86regsc = EDX;
    CheckReg(_Rt_, RM_R); CheckReg(_Rs_, RM_R);
	x86skip = 0;

	MOV32RtoR(EAX, regs[_Rt_]);
	MOV32RtoR(ECX, regs[_Rs_]);
	SHL32CLtoR(EAX);
	CheckReg(_Rd_, RM_W);
	MOV32RtoR(regs[_Rd_], EAX);
#else
	MOV32MtoR(EAX, (unsigned long)&cpuRegs.GPR.r[_Rt_]);
	MOV32MtoR(ECX, (unsigned long)&cpuRegs.GPR.r[_Rs_]);
	SHL32CLtoR(EAX);
	MOV32RtoM((unsigned long)&cpuRegs.GPR.r[_Rd_], EAX);
#endif
}

void recSRLV() {
// Rd = Rt >> Rs
	if (!_Rd_) return;

#ifdef CACHE_REGS
	FlushRegs();

	x86skip = ECX;
	x86regsc = EDX;
    CheckReg(_Rt_, RM_R); CheckReg(_Rs_, RM_R);
	x86skip = 0;

	MOV32RtoR(EAX, regs[_Rt_]);
	MOV32RtoR(ECX, regs[_Rs_]);
	SHR32CLtoR(EAX);
	CheckReg(_Rd_, RM_W); 
	MOV32RtoR(regs[_Rd_], EAX);
#else
	MOV32MtoR(EAX, (unsigned long)&cpuRegs.GPR.r[_Rt_]);
	MOV32MtoR(ECX, (unsigned long)&cpuRegs.GPR.r[_Rs_]);
	SHR32CLtoR(EAX);
	MOV32RtoM((unsigned long)&cpuRegs.GPR.r[_Rd_], EAX);
#endif
}

void recSRAV() {
// Rd = Rt >> Rs
	if (!_Rd_) return;

#ifdef CACHE_REGS
	FlushRegs();

	x86skip = ECX;
	x86regsc = EDX;
    CheckReg(_Rt_, RM_R); CheckReg(_Rs_, RM_R);
	x86skip = 0;

	MOV32RtoR(EAX, regs[_Rt_]);
	MOV32RtoR(ECX, regs[_Rs_]);
	SAR32CLtoR(EAX);
	CheckReg(_Rd_, RM_W);
	MOV32RtoR(regs[_Rd_], EAX);
#else
	MOV32MtoR(EAX, (unsigned long)&cpuRegs.GPR.r[_Rt_]);
	MOV32MtoR(ECX, (unsigned long)&cpuRegs.GPR.r[_Rs_]);
	SAR32CLtoR(EAX);
	MOV32RtoM((unsigned long)&cpuRegs.GPR.r[_Rd_], EAX);
#endif
}
//#endif

/*REC_SYS(SYSCALL);
REC_SYS(BREAK);

#if 0*/
void recSYSCALL() {
#ifdef CACHE_REGS
	FlushRegs();
#endif

	MOV32ItoR(EAX, pc - 4);
	MOV32RtoM((unsigned long)&cpuRegs.pc, EAX);
	PUSH32I  (branch == 1 ? 1 : 0);
	PUSH32I  (0x20);
	MOV32ItoR(EAX, (unsigned long)cpuTakeException);
	CALL32R  (EAX);
	ADD32ItoR(ESP, 8);

	if (!branch) branch = 2;
}

void recBREAK() {
#ifdef CACHE_REGS
	FlushRegs();
#endif

	MOV32ItoR(EAX, pc - 4);
	MOV32RtoM((unsigned long)&cpuRegs.pc, EAX);
	PUSH32I  (branch == 1 ? 1 : 0);
	PUSH32I  (0x40);
	MOV32ItoR(EAX, (unsigned long)cpuTakeException);
	CALL32R  (EAX);
	ADD32ItoR(ESP, 8);

	if (!branch) branch = 2;
}
//#endif
/*
REC_FUNC(MFHI);
REC_FUNC(MTHI);
REC_FUNC(MFLO);
REC_FUNC(MTLO);
#if 0*/
void recMFHI() {
// Rd = Hi
	if (!_Rd_) return;

#ifdef CACHE_REGS
	CheckReg(_Rd_, RM_W);

	MOV32MtoR(regs[_Rd_], (unsigned long)&cpuRegs.GPR.n.hi);
#else
	MOV32MtoR(EAX, (unsigned long)&cpuRegs.GPR.n.hi);
	MOV32RtoM((unsigned long)&cpuRegs.GPR.r[_Rd_], EAX);
#endif
}

void recMTHI() {
// Hi = Rs
#ifdef CACHE_REGS
	CheckReg(_Rs_, RM_R);

	MOV32RtoM((unsigned long)&cpuRegs.GPR.n.hi, regs[_Rs_]);
#else
	MOV32MtoR(EAX, (unsigned long)&cpuRegs.GPR.r[_Rs_]);
	MOV32RtoM((unsigned long)&cpuRegs.GPR.n.hi, EAX);
#endif
}

void recMFLO() {
// Rd = Lo
	if (!_Rd_) return;

#ifdef CACHE_REGS
	CheckReg(_Rd_, RM_W);

	MOV32MtoR(regs[_Rd_], (unsigned long)&cpuRegs.GPR.n.lo);
#else
	MOV32MtoR(EAX, (unsigned long)&cpuRegs.GPR.n.lo);
	MOV32RtoM((unsigned long)&cpuRegs.GPR.r[_Rd_], EAX);
#endif
}

void recMTLO() {
// Lo = Rs
#ifdef CACHE_REGS
	CheckReg(_Rs_, RM_R);

	MOV32RtoM((unsigned long)&cpuRegs.GPR.n.lo, regs[_Rs_]);
#else
	MOV32MtoR(EAX, (unsigned long)&cpuRegs.GPR.r[_Rs_]);
	MOV32RtoM((unsigned long)&cpuRegs.GPR.n.lo, EAX);
#endif
}
//#endif

/*REC_BRANCH(J);
REC_BRANCH(JR);
REC_BRANCH(JAL);
REC_BRANCH(JALR);
REC_BRANCH(BLTZ);
REC_BRANCH(BGTZ);
REC_BRANCH(BLTZAL);
REC_BRANCH(BGEZAL);
REC_BRANCH(BNE);
REC_BRANCH(BEQ);
REC_BRANCH(BLEZ);
REC_BRANCH(BGEZ);

#if 0*/
void recBLTZ() {
// Branch if Rs < 0
#ifdef CACHE_REGS
    CheckReg(_Rs_, RM_R);
#endif

	MOV32ItoR(EAX, _Imm_ * 4 + pc);
	MOV32RtoM((unsigned long)&target, EAX);
	MOV32ItoR(EAX, pc+4);
#ifdef CACHE_REGS
	CMP32ItoR(regs[_Rs_], 0);
#else
//	CMP32ItoM((unsigned long)&cpuRegs.GPR.r[_Rs_], 0);
	MOV32MtoR(ECX, (unsigned long)&cpuRegs.GPR.r[_Rs_]);
	CMP32ItoR(ECX, 0);
#endif
	CMOVL32MtoR(EAX, (unsigned long)&target);
	MOV32RtoM((unsigned long)&target, EAX);

	SetBranch();
}

void recBGEZ() {
// Branch if Rs >= 0
#ifdef CACHE_REGS
    CheckReg(_Rs_, RM_R);
#endif

	MOV32ItoR(EAX, _Imm_ * 4 + pc);
	MOV32RtoM((unsigned long)&target, EAX);
	MOV32ItoR(EAX, pc+4);
#ifdef CACHE_REGS
	CMP32ItoR(regs[_Rs_], 0);
#else
//	CMP32ItoM((unsigned long)&cpuRegs.GPR.r[_Rs_], 0);
	MOV32MtoR(ECX, (unsigned long)&cpuRegs.GPR.r[_Rs_]);
	CMP32ItoR(ECX, 0);
#endif
	CMOVGE32MtoR(EAX, (unsigned long)&target);
	MOV32RtoM((unsigned long)&target, EAX);

	SetBranch();
}

void recBLTZAL() {
// Branch if Rs < 0
#ifdef CACHE_REGS
    CheckReg(_Rs_, RM_R); CheckReg(31, RM_W);
#endif

	MOV32ItoR(EAX, _Imm_ * 4 + pc);
	MOV32RtoM((unsigned long)&target, EAX);
	MOV32ItoR(EAX, pc+4);
#ifdef CACHE_REGS
	CMP32ItoR(regs[_Rs_], 0);
#else
//	CMP32ItoM((unsigned long)&cpuRegs.GPR.r[_Rs_], 0);
	MOV32MtoR(ECX, (unsigned long)&cpuRegs.GPR.r[_Rs_]);
	CMP32ItoR(ECX, 0);
#endif
	CMOVL32MtoR(EAX, (unsigned long)&target);
	MOV32RtoM((unsigned long)&target, EAX);

#ifdef CACHE_REGS
	MOV32ItoR(regs[31], pc + 4);
#else
	MOV32ItoM((unsigned long)&cpuRegs.GPR.r[31], pc + 4);
#endif

	SetBranch();
}

void recBGEZAL() {
// Branch if Rs >= 0
#ifdef CACHE_REGS
    CheckReg(_Rs_, RM_R); CheckReg(31, RM_W);
#endif

	MOV32ItoR(EAX, _Imm_ * 4 + pc);
	MOV32RtoM((unsigned long)&target, EAX);
	MOV32ItoR(EAX, pc+4);
#ifdef CACHE_REGS
	CMP32ItoR(regs[_Rs_], 0);
#else
	MOV32MtoR(ECX, (unsigned long)&cpuRegs.GPR.r[_Rs_]);
	CMP32ItoR(ECX, 0);
//	CMP32ItoM((unsigned long)&cpuRegs.GPR.r[_Rs_], 0);
#endif
	CMOVGE32MtoR(EAX, (unsigned long)&target);
	MOV32RtoM((unsigned long)&target, EAX);

#ifdef CACHE_REGS
	MOV32ItoR(regs[31], pc + 4);
#else
	MOV32ItoM((unsigned long)&cpuRegs.GPR.r[31], pc + 4);
#endif

	SetBranch();
}

void recJ() {
// j target
	MOV32ItoM((unsigned long)&target, _Target_ * 4 + (pc & 0xf0000000));

	SetBranch();
}

void recJR() {
// jr Rs
#ifdef CACHE_REGS
    CheckReg(_Rs_, RM_R);

	MOV32RtoM((unsigned long)&target, regs[_Rs_]);
#else
	MOV32MtoR(EAX, (unsigned long)&cpuRegs.GPR.r[_Rs_]);
	MOV32RtoM((unsigned long)&target, EAX);
#endif

	SetBranch();
}

void recJALR() {
// jalr Rs
#ifdef CACHE_REGS
    CheckReg(_Rs_, RM_R); CheckReg(31, RM_W);

	MOV32RtoM((unsigned long)&target, regs[_Rs_]);

	MOV32ItoR(regs[31], pc + 4);
#else
	MOV32MtoR(EAX, (unsigned long)&cpuRegs.GPR.r[_Rs_]);
	MOV32RtoM((unsigned long)&target, EAX);

	MOV32ItoM((unsigned long)&cpuRegs.GPR.r[31], pc + 4);
#endif
	
	SetBranch();
}

void recJAL() {
// jal target
#ifdef CACHE_REGS
	CheckReg(31, RM_W);
#endif

	MOV32ItoM((unsigned long)&target, _Target_ * 4 + (pc & 0xf0000000));

#ifdef CACHE_REGS
	MOV32ItoR(regs[31], pc + 4);
#else
	MOV32ItoM((unsigned long)&cpuRegs.GPR.r[31], pc + 4);
#endif

	SetBranch();
}

void recBEQ() {
// Branch if Rs == Rt
#ifdef CACHE_REGS
    CheckReg(_Rt_, RM_R); CheckReg(_Rs_, RM_R);
#endif

	MOV32ItoR(EAX, _Imm_ * 4 + pc);
	MOV32RtoM((unsigned long)&target, EAX);
	MOV32ItoR(EAX, pc+4);
#ifdef CACHE_REGS
	CMP32RtoR(regs[_Rs_], regs[_Rt_]);
#else
	MOV32MtoR(ECX, (unsigned long)&cpuRegs.GPR.r[_Rs_]);
	MOV32MtoR(EDX, (unsigned long)&cpuRegs.GPR.r[_Rt_]);
	CMP32RtoR(ECX, EDX);
//	CMP32MtoR(EAX, (unsigned long)&cpuRegs.GPR.r[_Rt_]);
#endif
	CMOVE32MtoR(EAX, (unsigned long)&target);
	MOV32RtoM((unsigned long)&target, EAX);

	SetBranch();
}

void recBNE() {
// Branch if Rs != Rt
#ifdef CACHE_REGS
    CheckReg(_Rt_, RM_R); CheckReg(_Rs_, RM_R);
#endif

	MOV32ItoR(EAX, _Imm_ * 4 + pc);
	MOV32RtoM((unsigned long)&target, EAX);
	MOV32ItoR(EAX, pc+4);
#ifdef CACHE_REGS
	CMP32RtoR(regs[_Rs_], regs[_Rt_]);
#else
	MOV32MtoR(ECX, (unsigned long)&cpuRegs.GPR.r[_Rs_]);
	MOV32MtoR(EDX, (unsigned long)&cpuRegs.GPR.r[_Rt_]);
	CMP32RtoR(ECX, EDX);
//	CMP32MtoR(EAX, (unsigned long)&cpuRegs.GPR.r[_Rt_]);
#endif
	CMOVNE32MtoR(EAX, (unsigned long)&target);
	MOV32RtoM((unsigned long)&target, EAX);

	SetBranch();
}

void recBLEZ() {
// Branch if Rs <= 0
#ifdef CACHE_REGS
    CheckReg(_Rs_, RM_R);
#endif

	MOV32ItoR(EAX, _Imm_ * 4 + pc);
	MOV32RtoM((unsigned long)&target, EAX);
	MOV32ItoR(EAX, pc+4);
#ifdef CACHE_REGS
	CMP32ItoR(regs[_Rs_], 0);
#else
	MOV32MtoR(ECX, (unsigned long)&cpuRegs.GPR.r[_Rs_]);
	CMP32ItoR(ECX, 0);
//	CMP32ItoM((unsigned long)&cpuRegs.GPR.r[_Rs_], 0);
#endif
	CMOVLE32MtoR(EAX, (unsigned long)&target);
	MOV32RtoM((unsigned long)&target, EAX);

	SetBranch();
}

void recBGTZ() {
// Branch if Rs > 0
#ifdef CACHE_REGS
    CheckReg(_Rs_, RM_R);
#endif

	MOV32ItoR(EAX, _Imm_ * 4 + pc);
	MOV32RtoM((unsigned long)&target, EAX);
	MOV32ItoR(EAX, pc+4);
#ifdef CACHE_REGS
	CMP32ItoR(regs[_Rs_], 0);
#else
	MOV32MtoR(ECX, (unsigned long)&cpuRegs.GPR.r[_Rs_]);
	CMP32ItoR(ECX, 0);
//	CMP32ItoM((unsigned long)&cpuRegs.GPR.r[_Rs_], 0);
#endif
	CMOVG32MtoR(EAX, (unsigned long)&target);
	MOV32RtoM((unsigned long)&target, EAX);

	SetBranch();
}
//#endif

/*REC_FUNC(MFC0);
REC_FUNC(MTC0);
REC_FUNC(CFC0);
REC_FUNC(CTC0);
REC_FUNC(RFE);
#if 0*/
void recMFC0() {
// Rt = Cop0->Rd
	if (!_Rt_) return;

#ifdef CACHE_REGS
	CheckReg(_Rt_, RM_W);

	MOV32MtoR(regs[_Rt_], (unsigned long)&cpuRegs.CP0.r[_Rd_]);
#else
	MOV32MtoR(EAX, (unsigned long)&cpuRegs.CP0.r[_Rd_]);
	MOV32RtoM((unsigned long)&cpuRegs.GPR.r[_Rt_], EAX);
#endif
}

void recCFC0() {
// Rt = Cop0->Rd
	if (!_Rt_) return;

#ifdef CACHE_REGS
	CheckReg(_Rt_, RM_W);

	MOV32MtoR(regs[_Rt_], (unsigned long)&cpuRegs.CP0.r[_Rd_]);
#else
	MOV32MtoR(EAX, (unsigned long)&cpuRegs.CP0.r[_Rd_]);
	MOV32RtoM((unsigned long)&cpuRegs.GPR.r[_Rt_], EAX);
#endif
}

void recMTC0() {
// Cop0->Rd = Rt
#ifdef CACHE_REGS
	CheckReg(_Rt_, RM_R);

	MOV32RtoM((unsigned long)&cpuRegs.CP0.r[_Rd_], regs[_Rt_]);
#else
	MOV32MtoR(EAX, (unsigned long)&cpuRegs.GPR.r[_Rt_]);
	MOV32RtoM((unsigned long)&cpuRegs.CP0.r[_Rd_], EAX);
#endif
}

void recCTC0() {
// Cop0->Rd = Rt
#ifdef CACHE_REGS
	CheckReg(_Rt_, RM_R);

	MOV32RtoM((unsigned long)&cpuRegs.CP0.r[_Rd_], regs[_Rt_]);
#else
	MOV32MtoR(EAX, (unsigned long)&cpuRegs.GPR.r[_Rt_]);
	MOV32RtoM((unsigned long)&cpuRegs.CP0.r[_Rd_], EAX);
#endif
}

void recRFE() {
#ifdef CACHE_REGS
	FlushRegs();
//	ClearX86Reg(ECX);
#endif
	MOV32MtoR(EAX, (unsigned long)&cpuRegs.CP0.n.Status);
	MOV32RtoR(ECX, EAX);
	AND32ItoR(EAX, 0xfffffff0);
	AND32ItoR(ECX, 0x3c);
	SHR32ItoR(ECX, 2);
	OR32RtoR (EAX, ECX);
	MOV32RtoM((unsigned long)&cpuRegs.CP0.n.Status, EAX);
}
//#endif

#define CP2_FUNC(f) \
void cp2##f(); \
void rec##f() { \
	FlushRegs(); \
 \
	MOV32ItoR(EAX, (unsigned long)cpuRegs.code); \
	MOV32RtoM((unsigned long)&cpuRegs.code, EAX); \
	MOV32ItoR(EAX, (unsigned long)pc); \
	MOV32RtoM((unsigned long)&cpuRegs.pc, EAX); \
	MOV32ItoR(EAX, (unsigned long)cp2##f); \
	CALL32R  (EAX); \
	MOV32ItoR(EAX, 0); \
	MOV32RtoM((unsigned long)&cpuRegs.GPR.n.r0, EAX); \
/*	branch = 2; */\
}

CP2_FUNC(MFC2);
CP2_FUNC(MTC2);
CP2_FUNC(CFC2);
CP2_FUNC(CTC2);
CP2_FUNC(LWC2);
CP2_FUNC(SWC2);

CP2_FUNC(RTPS);
CP2_FUNC(OP);
CP2_FUNC(NCLIP);
CP2_FUNC(DPCS);
CP2_FUNC(INTPL);
CP2_FUNC(MVMVA);
CP2_FUNC(NCDS);
CP2_FUNC(NCDT);
CP2_FUNC(CDP);
CP2_FUNC(NCCS);
CP2_FUNC(CC);
CP2_FUNC(NCS);
CP2_FUNC(NCT);
CP2_FUNC(SQR);
CP2_FUNC(DCPL);
CP2_FUNC(DPCT);
CP2_FUNC(AVSZ3);
CP2_FUNC(AVSZ4);
CP2_FUNC(RTPT);
CP2_FUNC(GPF);
CP2_FUNC(GPL);
CP2_FUNC(NCCT);

//

static void (*recBSC[64])() = {
	recSPECIAL, recREGIMM, recJ   , recJAL  , recBEQ , recBNE , recBLEZ, recBGTZ,
	recADDI   , recADDIU , recSLTI, recSLTIU, recANDI, recORI , recXORI, recLUI ,
	recCOP0   , recNULL  , recCOP2, recNULL , recNULL, recNULL, recNULL, recNULL,
	recNULL   , recNULL  , recNULL, recNULL , recNULL, recNULL, recNULL, recNULL,
	recLB     , recLH    , recLWL , recLW   , recLBU , recLHU , recLWR , recNULL,
	recSB     , recSH    , recSWL , recSW   , recNULL, recNULL, recSWR , recNULL,
	recNULL   , recNULL  , recLWC2, recNULL , recNULL, recNULL, recNULL, recNULL,
	recNULL   , recNULL  , recSWC2, recNULL , recNULL, recNULL, recNULL, recNULL
};

static void (*recSPC[64])() = {
	recSLL , recNULL, recSRL , recSRA , recSLLV   , recNULL , recSRLV, recSRAV,
	recJR  , recJALR, recNULL, recNULL, recSYSCALL, recBREAK, recNULL, recNULL,
	recMFHI, recMTHI, recMFLO, recMTLO, recNULL   , recNULL , recNULL, recNULL,
	recMULT, recMULTU, recDIV, recDIVU, recNULL   , recNULL , recNULL, recNULL,
	recADD , recADDU, recSUB , recSUBU, recAND    , recOR   , recXOR , recNOR ,
	recNULL, recNULL, recSLT , recSLTU, recNULL   , recNULL , recNULL, recNULL,
	recNULL, recNULL, recNULL, recNULL, recNULL   , recNULL , recNULL, recNULL,
	recNULL, recNULL, recNULL, recNULL, recNULL   , recNULL , recNULL, recNULL
};

static void (*recREG[32])() = {
	recBLTZ  , recBGEZ  , recNULL, recNULL, recNULL, recNULL, recNULL, recNULL,
	recNULL  , recNULL  , recNULL, recNULL, recNULL, recNULL, recNULL, recNULL,
	recBLTZAL, recBGEZAL, recNULL, recNULL, recNULL, recNULL, recNULL, recNULL,
	recNULL  , recNULL  , recNULL, recNULL, recNULL, recNULL, recNULL, recNULL
};

static void (*recCP0[32])() = {
	recMFC0, recNULL, recCFC0, recNULL, recMTC0, recNULL, recCTC0, recNULL,
	recNULL, recNULL, recNULL, recNULL, recNULL, recNULL, recNULL, recNULL,
	recRFE , recNULL, recNULL, recNULL, recNULL, recNULL, recNULL, recNULL,
	recNULL, recNULL, recNULL, recNULL, recNULL, recNULL, recNULL, recNULL
};

static void (*recCP2[64])() = {
	recBASIC, recRTPS , recNULL , recNULL, recNULL, recNULL , recNCLIP, recNULL, // 00
	recNULL , recNULL , recNULL , recNULL, recOP  , recNULL , recNULL , recNULL, // 08
	recDPCS , recINTPL, recMVMVA, recNCDS, recCDP , recNULL , recNCDT , recNULL, // 10
	recNULL , recNULL , recNULL , recNCCS, recCC  , recNULL , recNCS  , recNULL, // 18
	recNCT  , recNULL , recNULL , recNULL, recNULL, recNULL , recNULL , recNULL, // 20
	recSQR  , recDCPL , recDPCT , recNULL, recNULL, recAVSZ3, recAVSZ4, recNULL, // 28 
	recRTPT , recNULL , recNULL , recNULL, recNULL, recNULL , recNULL , recNULL, // 30
	recNULL , recNULL , recNULL , recNULL, recNULL, recGPF  , recGPL  , recNCCT  // 38
};

static void (*recCP2BSC[32])() = {
	recMFC2, recNULL, recCFC2, recNULL, recMTC2, recNULL, recCTC2, recNULL,
	recNULL, recNULL, recNULL, recNULL, recNULL, recNULL, recNULL, recNULL,
	recNULL, recNULL, recNULL, recNULL, recNULL, recNULL, recNULL, recNULL,
	recNULL, recNULL, recNULL, recNULL, recNULL, recNULL, recNULL, recNULL
};

static void recRecompile() {
	/* if recPtr reached the mem limit reset whole mem */
	if (((unsigned long)recPtr - (unsigned long)recMem) >= (0x800000 - 0x8000))
		recReset();

	PC_REC32(cpuRegs.pc) = (unsigned long)recPtr;
	pc = cpuRegs.pc;

#ifdef CACHE_REGS
	PUSHA32();
#endif

	for (count=0; count<500;) {
		char *p;

		p = (char *)PC_MEM(pc);
		if (p == NULL) recError();
		cpuRegs.code = *(unsigned long *)p;

		pc+=4;
		count++;
//printf ("%s\n", disR3000AF(cpuRegs.code, pc));
//FlushRegs();
		recBSC[cpuRegs.code>>26]();

		if (branch) {
			break;
		}
	}

	/* update cpu regs */
	FlushRegs();

	if (!branch) {
		MOV32ItoR(EAX, pc);
		MOV32RtoM((unsigned long)&cpuRegs.pc, EAX);
	}

	/* store cycle */
	MOV32MtoR(EAX, (unsigned long)&cpuIrq.cycle);
	ADD32ItoR(EAX, count);
	MOV32RtoM((unsigned long)&cpuIrq.cycle, EAX);

	branch = 0;

#ifdef CACHE_REGS
	POPA32();
#endif

	if (mode) {
		RET();
	} else {
		MOV32ItoR(EAX, (unsigned long)recMem);
		JMP32R(EAX);
	}
}


R3000Acpu recCpu = {
	recInit,
	recReset,
	recExecuteBios,
	recExecute,
	recClear,
	recShutdown
};
