/*  Pcsx - Pc Psx Emulator
 *  Copyright (C) 1999-2002  Pcsx Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "Debug.h"

long jumpMode;
char ostr[256];

// Names of registers
char *disRNameGPR[] = {
	"r0", "at", "v0", "v1", "a0", "a1","a2", "a3",
	"t0", "t1", "t2", "t3", "t4", "t5","t6", "t7",
	"s0", "s1", "s2", "s3", "s4", "s5","s6", "s7",
	"t8", "t9", "k0", "k1", "gp", "sp","fp", "ra"};

char *disRNameCOP0[] = {
	"Index"     , "Random"    , "EntryLo0", "EntryLo1", "Context" , "PageMask"  , "Wired"     , "*Check me*",
	"BadVAddr"  , "Count"     , "EntryHi" , "Compare" , "Status"  , "Cause"     , "ExceptPC"  , "PRevID"    ,
	"Config"    , "LLAddr"    , "WatchLo" , "WatchHi" , "XContext", "*RES*"     , "*RES*"     , "*RES*"     ,
	"*RES*"     , "*RES* "    , "PErr"    , "CacheErr", "TagLo"   , "TagHi"     , "ErrorEPC"  , "*RES*"     };


// Type deffinition of our functions
#define DisFInterface  (unsigned long code, unsigned long pc)
#define DisFInterfaceT (unsigned long, unsigned long)
#define DisFInterfaceN (code, pc)

typedef char* (*TdisR3000AF)DisFInterface;

// These macros are used to assemble the disassembler functions
#define MakeDisF(fn, b) char* fn DisFInterface { b; return ostr; }


#include "R3000A.h"
#undef _Target_
#undef _Branch_
#undef _Funct_
#undef _Rd_
#undef _Rt_
#undef _Rs_
#undef _Sa_
#undef _Im_

#define _Funct_  ((code      ) & 0x3F) // The funct part of the instruction register 
#define _Rd_     ((code >> 11) & 0x1F) // The rd part of the instruction register 
#define _Rt_     ((code >> 16) & 0x1F) // The rt part of the instruction register 
#define _Rs_     ((code >> 21) & 0x1F) // The rs part of the instruction register 
#define _Sa_     ((code >>  6) & 0x1F) // The sa part of the instruction register
#define _Im_     ( code & 0xFFFF)      // The immediate part of the instruction register

#define DispDis(x) sprintf (ostr, "%8.8lx %8.8lx:", pc, code); x

#define _nRs_ _u32(_rRs_), disRNameGPR[_Rs_]
#define _nRt_ _u32(_rRt_), disRNameGPR[_Rt_]
#define _nRd_ _u32(_rRd_), disRNameGPR[_Rd_]
#define _nSa_ _u32(_rSa_), disRNameGPR[_Sa_]

#define _nRd0_ _u32(_rRd_), disRNameCOP0[_Rd_]

#define _vSa_ _Sa_
#define _vRt_ _Rt_

#define _I_       _Im_, _Im_
#define _Target_  ((pc & 0xf0000000) + (code & 0x03ffffff * 4))
#define _Branch_  (pc + 4 + ((short)_Im_ * 4))
#define _OfB_     _Im_, _nRs_

// These macros represent the different instruction formats
#define DispITypeRsRtO(i)  DispDis(sprintf (ostr, "%s %-7s, %8.8lx (%s), %8.8lx (%s), %8.8lx", ostr, i, _nRs_, _nRt_, _Branch_))
#define DispITypeRtRsI(i)  DispDis(sprintf (ostr, "%s %-7s, %8.8lx (%s), %8.8lx (%s), %4.4lx (%ld)", ostr, i, _nRt_, _nRs_, _I_))
#define DispITypeRtOB(i)   DispDis(sprintf (ostr, "%s %-7s, %8.8lx (%s), %8.8lx (%8.8lx (%s))", ostr, i, _nRt_, _OfB_))
#define DispITypeRtOB1(i)  DispDis(sprintf (ostr, "%s %-7s, %8.8lx (%s), %8.8lx (%8.8lx (%s))", ostr, i, _nFt_, _OfB_))
#define DispITypeOpOB(i)   DispDis(sprintf (ostr, "%s %-7s, %8.8lx, %x (%4.4lx (%s))", ostr, i, _vRt_, _OfB_))
#define DispITypeRsO(i)    DispDis(sprintf (ostr, "%s %-7s, %8.8lx (%s), %8.8lx", ostr, i, _nRs_, _Branch_))
#define DispITypeRsI(i)    DispDis(sprintf (ostr, "%s %-7s, %8.8lx (%s), %4.4lx (%ld)", ostr, i, _nRs_, _I_))
#define DispITypeRtI(i)    DispDis(sprintf (ostr, "%s %-7s, %8.8lx (%s), %4.4lx (%ld)", ostr, i, _nRt_, _I_))
#define DispITypeRd(i)     DispDis(sprintf (ostr, "%s %-7s, %8.8lx (%s)", ostr, i, _nRd_))
#define DispITypeRs(i)     DispDis(sprintf (ostr, "%s %-7s, %8.8lx (%s)", ostr, i, _nRs_))
#define DispITypeO(i)      DispDis(sprintf (ostr, "%s %-7s, %8.8lx", ostr, i, _Branch_))
#define DispIType(i)       DispDis(sprintf (ostr, "%s %-7s", ostr, i))

#define DispRTypeRdRtRs(i) DispDis(sprintf (ostr, "%s %-7s, %8.8lx (%s), %8.8lx (%s), %8.8lx (%s)", ostr, i, _nRd_, _nRt_, _nRs_))
#define DispRTypeRdRsRt(i) DispDis(sprintf (ostr, "%s %-7s, %8.8lx (%s), %8.8lx (%s), %8.8lx (%s)", ostr, i, _nRd_, _nRs_, _nRt_))
#define DispRTypeRdRtSa(i) DispDis(sprintf (ostr, "%s %-7s, %8.8lx (%s), %8.8lx (%s), %lx (%ld)"   , ostr, i, _nRd_, _nRt_, _vSa_, _vSa_))
#define DispRTypeRtRd0(i)  DispDis(sprintf (ostr, "%s %-7s, %8.8lx (%s), %8.8lx (%s)", ostr, i, _nRt_, _nRd0_))
#define DispRTypeRtRd1(i)  DispDis(sprintf (ostr, "%s %-7s, %8.8lx (%s), %8.8lx (%s)", ostr, i, _nRt_, _nRd1_))
#define DispRTypeRsRd(i)   DispDis(sprintf (ostr, "%s %-7s, %8.8lx (%s), %8.8lx (%s)", ostr, i, _nRs_, _nRd_))
#define DispRTypeRsRt(i)   DispDis(sprintf (ostr, "%s %-7s, %8.8lx (%s), %8.8lx (%s)", ostr, i, _nRs_, _nRt_))
#define DispRTypeRtRd(i)   DispDis(sprintf (ostr, "%s %-7s, %8.8lx (%s), %8.8lx (%s)", ostr, i, _nRt_, _nRd_))

#define DispFTypeFdFsFt(i) DispDis(sprintf (ostr, "%s %-7s, %8.8lx (%s), %8.8lx (%s), %8.8lx (%s)", ostr, i, _nFd_, _nFs_, _nFt_))
#define DispFTypeFdFs(i)   DispDis(sprintf (ostr, "%s %-7s, %8.8lx (%s), %8.8lx (%s)", ostr, i, _nFd_, _nFs_))
#define DispFTypeFsFt(i)   DispDis(sprintf (ostr, "%s %-7s, %8.8lx (%s), %8.8lx (%s)", ostr, i, _nFs_, _nFt_))

#define DispJType(i)       DispDis(sprintf (ostr, "%s %-7s, %8.8lx", ostr, i, _Target_))

#define DispCall(i)        DispDis(sprintf (ostr, "%s Call %s", ostr, i))


/*********************************************************
* Arithmetic with immediate operand                      *
* Format:  OP rt, rs, immediate                          *
*********************************************************/
MakeDisF(disADDI,		DispITypeRtRsI("ADDI"))
MakeDisF(disADDIU,		DispITypeRtRsI("ADDIU"))
MakeDisF(disANDI,		DispITypeRtRsI("ANDI"))
MakeDisF(disORI,		DispITypeRtRsI("ORI"))
MakeDisF(disSLTI,		DispITypeRtRsI("SLTI"))
MakeDisF(disSLTIU,		DispITypeRtRsI("SLTIU"))
MakeDisF(disXORI,		DispITypeRtRsI("XORI"))

/*********************************************************
* Register arithmetic                                    *
* Format:  OP rd, rs, rt                                 *
*********************************************************/
MakeDisF(disADD,		DispRTypeRdRsRt("ADD"))
MakeDisF(disADDU,		DispRTypeRdRsRt("ADDU"))
MakeDisF(disAND,		DispRTypeRdRsRt("AND"))
MakeDisF(disNOR,		DispRTypeRdRsRt("NOR"))
MakeDisF(disOR,			DispRTypeRdRsRt("OR"))
MakeDisF(disSLT,		DispRTypeRdRsRt("SLT"))
MakeDisF(disSLTU,		DispRTypeRdRsRt("SLTU"))
MakeDisF(disSUB,		DispRTypeRdRsRt("SUB"))
MakeDisF(disSUBU,		DispRTypeRdRsRt("SUBU"))
MakeDisF(disXOR,		DispRTypeRdRsRt("XOR"))

/*********************************************************
* Register arithmetic & Register trap logic              *
* Format:  OP rs, rt                                     *
*********************************************************/
MakeDisF(disDIV,		DispRTypeRsRt("DIV"))
MakeDisF(disDIVU,		DispRTypeRsRt("DIVU"))
MakeDisF(disMULT,		DispRTypeRsRt("MULT"))
MakeDisF(disMULTU,		DispRTypeRsRt("MULTU"))

/*********************************************************
* Register branch logic                                  *
* Format:  OP rs, offset                                 *
*********************************************************/
MakeDisF(disBGEZ,		jumpMode = 3; DispITypeRsO("BGEZ"))
MakeDisF(disBGEZAL,		jumpMode = 3; DispITypeRsO("BGEZAL"))
MakeDisF(disBGTZ,		jumpMode = 3; DispITypeRsO("BGTZ"))
MakeDisF(disBLEZ,		jumpMode = 3; DispITypeRsO("BLEZ"))
MakeDisF(disBLTZ,		jumpMode = 3; DispITypeRsO("BLTZ"))
MakeDisF(disBLTZAL,		jumpMode = 3; DispITypeRsO("BLTZAL"))

/*********************************************************
* Shift arithmetic with constant shift                   *
* Format:  OP rd, rt, sa                                 *
*********************************************************/
MakeDisF(disSLL,		if (code) { DispRTypeRdRtSa("SLL"); } else { DispDis(sprintf (ostr, "%s NOP", ostr);) })
MakeDisF(disSRA,		DispRTypeRdRtSa("SRA"))
MakeDisF(disSRL,		DispRTypeRdRtSa("SRL"))

/*********************************************************
* Shift arithmetic with variant register shift           *
* Format:  OP rd, rt, rs                                 *
*********************************************************/
MakeDisF(disSLLV,		DispRTypeRdRtRs("SLLV"))
MakeDisF(disSRAV,		DispRTypeRdRtRs("SRAV"))
MakeDisF(disSRLV,		DispRTypeRdRtRs("SRLV"))

/*********************************************************
* Load higher 16 bits of the first word in GPR with imm  *
* Format:  OP rt, immediate                              *
*********************************************************/
MakeDisF(disLUI,		DispITypeRtI("LUI"))

/*********************************************************
* Move from HI/LO to GPR                                 *
* Format:  OP rd                                         *
*********************************************************/
MakeDisF(disMFHI,		DispITypeRd("MFHI"))
MakeDisF(disMFLO,		DispITypeRd("MFLO"))

/*********************************************************
* Move from GPR to HI/LO & Register jump                 *
* Format:  OP rd                                         *
*********************************************************/
MakeDisF(disMTHI,		DispITypeRs("MTHI"))
MakeDisF(disMTLO,		DispITypeRs("MTLO"))
MakeDisF(disJR,			jumpMode = 3; DispITypeRs("JR"))

/*********************************************************
* Special purpose instructions                           *
* Format:  OP                                            *
*********************************************************/
MakeDisF(disBREAK,		DispIType("BREAK"))
MakeDisF(disRFE,		DispIType("RFE"))
MakeDisF(disSYSCALL,	DispIType("SYSCALL"))


MakeDisF(disRTPS,		DispIType("RTPS"))
MakeDisF(disOP  ,		DispIType("OP"))
MakeDisF(disNCLIP,		DispIType("NCLIP"))
MakeDisF(disDPCS,		DispIType("DPCS"))
MakeDisF(disINTPL,		DispIType("INTPL"))
MakeDisF(disMVMVA,		DispIType("MVMVA"))
MakeDisF(disNCDS ,		DispIType("NCDS"))
MakeDisF(disCDP ,		DispIType("CDP"))
MakeDisF(disNCDT ,		DispIType("NCDT"))
MakeDisF(disNCCS ,		DispIType("NCCS"))
MakeDisF(disCC  ,		DispIType("CC"))
MakeDisF(disNCS ,		DispIType("NCS"))
MakeDisF(disNCT  ,		DispIType("NCT"))
MakeDisF(disSQR  ,		DispIType("SQR"))
MakeDisF(disDCPL ,		DispIType("DCPL"))
MakeDisF(disDPCT ,		DispIType("DPCT"))
MakeDisF(disAVSZ3,		DispIType("AVSZ3"))
MakeDisF(disAVSZ4,		DispIType("AVSZ4"))
MakeDisF(disRTPT ,		DispIType("RTPT"))
MakeDisF(disGPF  ,		DispIType("GPF"))
MakeDisF(disGPL  ,		DispIType("GPL"))
MakeDisF(disNCCT ,		DispIType("NCCT"))

MakeDisF(disMFC2,		DispIType("MFC2"))
MakeDisF(disCFC2,		DispIType("CFC2"))
MakeDisF(disMTC2,		DispIType("MTC2"))
MakeDisF(disCTC2,		DispIType("CTC2"))

/*********************************************************
* Register branch logic                                  *
* Format:  OP rs, rt, offset                             *
*********************************************************/
MakeDisF(disBEQ,		jumpMode = 3; DispITypeRsRtO("BEQ"))
MakeDisF(disBNE,		jumpMode = 3; DispITypeRsRtO("BNE"))

/*********************************************************
* Jump to target                                         *
* Format:  OP target                                     *
*********************************************************/
MakeDisF(disJ,			jumpMode = 3; DispJType("J"))
MakeDisF(disJAL,		jumpMode = 2; DispJType("JAL"))

/*********************************************************
* Register jump                                          *
* Format:  OP rs, rd                                     *
*********************************************************/
MakeDisF(disJALR,		jumpMode = 3; DispRTypeRsRd("JALR"))

/*********************************************************
* Load and store for GPR                                 *
* Format:  OP rt, offset(base)                           *
*********************************************************/
MakeDisF(disLB,			DispITypeRtOB("LB"))
MakeDisF(disLBU,		DispITypeRtOB("LBU"))
MakeDisF(disLH,			DispITypeRtOB("LH"))
MakeDisF(disLHU,		DispITypeRtOB("LHU"))
MakeDisF(disLW,			DispITypeRtOB("LW"))
MakeDisF(disLWL,		DispITypeRtOB("LWL"))
MakeDisF(disLWR,		DispITypeRtOB("LWR"))
MakeDisF(disLWC2,		DispITypeRtOB("LWC2"))
MakeDisF(disSB,			DispITypeRtOB("SB"))
MakeDisF(disSH,			DispITypeRtOB("SH"))
MakeDisF(disSW,			DispITypeRtOB("SW"))
MakeDisF(disSWL,		DispITypeRtOB("SWL"))
MakeDisF(disSWR,		DispITypeRtOB("SWR"))
MakeDisF(disSWC2,		DispITypeRtOB("SWC2"))

/*********************************************************
* Moves between GPR and COPx                             *
* Format:  OP rt, fs                                     *
*********************************************************/
MakeDisF(disMFC0,		DispRTypeRtRd0("MFC0"))
MakeDisF(disMTC0,		DispRTypeRtRd0("MTC0"))
MakeDisF(disCFC0,		DispRTypeRtRd0("CFC0"))
MakeDisF(disCTC0,		DispRTypeRtRd0("CTC0"))

/*********************************************************
* Unknow instruction (would generate an exception)       *
* Format:  ?                                             *
*********************************************************/
MakeDisF(disNULL,		DispDis(sprintf (ostr, "%s *** Bad OP ***", ostr)))


TdisR3000AF disR3000A_SPECIAL[] = { // Subset of disSPECIAL
	disSLL , disNULL , disSRL , disSRA , disSLLV   , disNULL  , disSRLV  , disSRAV ,
	disJR  , disJALR , disNULL, disNULL, disSYSCALL, disBREAK , disNULL  , disNULL ,
	disMFHI, disMTHI , disMFLO, disMTLO, disNULL   , disNULL  , disNULL  , disNULL ,
	disMULT, disMULTU, disDIV , disDIVU, disNULL   , disNULL  , disNULL  , disNULL ,
	disADD , disADDU , disSUB , disSUBU, disAND    , disOR    , disXOR   , disNOR  ,
	disNULL, disNULL , disSLT , disSLTU, disNULL   , disNULL  , disNULL  , disNULL ,
	disNULL, disNULL , disNULL, disNULL, disNULL   , disNULL  , disNULL   , disNULL ,
	disNULL, disNULL , disNULL, disNULL, disNULL   , disNULL  , disNULL  , disNULL};

MakeDisF(disSPECIAL,	disR3000A_SPECIAL[_Funct_] DisFInterfaceN)

TdisR3000AF disR3000A_BCOND[] = { // Subset of disBCOND
	disBLTZ  , disBGEZ  , disNULL, disNULL, disNULL, disNULL, disNULL, disNULL,
	disNULL  , disNULL  , disNULL, disNULL, disNULL, disNULL, disNULL, disNULL,
	disBLTZAL, disBGEZAL, disNULL, disNULL, disNULL, disNULL, disNULL, disNULL,
	disNULL  , disNULL  , disNULL, disNULL, disNULL, disNULL, disNULL, disNULL};

MakeDisF(disBCOND,	disR3000A_BCOND[_Rt_] DisFInterfaceN)

TdisR3000AF disR3000A_COP0[] = { // Subset of disCOP0
	disMFC0, disNULL, disCFC0, disNULL, disMTC0, disNULL, disCTC0, disNULL,
	disNULL, disNULL, disNULL, disNULL, disNULL, disNULL, disNULL, disNULL,
	disRFE , disNULL, disNULL, disNULL, disNULL, disNULL, disNULL, disNULL,
	disNULL, disNULL, disNULL, disNULL, disNULL, disNULL, disNULL, disNULL};

MakeDisF(disCOP0,		disR3000A_COP0[_Rs_] DisFInterfaceN)

TdisR3000AF disR3000A_BASIC[] = { // Subset of disBASIC (based on rs)
	disMFC2, disNULL, disCFC2, disNULL, disMTC2, disNULL, disCTC2, disNULL,
	disNULL, disNULL, disNULL, disNULL, disNULL, disNULL, disNULL, disNULL,
	disNULL, disNULL, disNULL, disNULL, disNULL, disNULL, disNULL, disNULL,
	disNULL, disNULL, disNULL, disNULL, disNULL, disNULL, disNULL, disNULL};

MakeDisF(disBASIC,		disR3000A_BASIC[_Rs_] DisFInterfaceN)

TdisR3000AF disR3000A_COP2[] = { // Subset of disR3000F_COP2 (based on funct)
	disBASIC, disRTPS , disNULL , disNULL , disNULL, disNULL , disNCLIP, disNULL,
	disNULL , disNULL , disNULL , disNULL , disOP  , disNULL , disNULL , disNULL,
	disDPCS , disINTPL, disMVMVA, disNCDS , disCDP , disNULL , disNCDT , disNULL,
	disNULL , disNULL , disNULL , disNCCS , disCC  , disNULL , disNCS  , disNULL,
	disNCT  , disNULL , disNULL , disNULL , disNULL, disNULL , disNULL , disNULL,
	disSQR  , disDCPL , disDPCT , disNULL , disNULL, disAVSZ3, disAVSZ4, disNULL, 
	disRTPT , disNULL , disNULL , disNULL , disNULL, disNULL , disNULL , disNULL,
	disNULL , disNULL , disNULL , disNULL , disNULL, disGPF  , disGPL  , disNCCT   };

MakeDisF(disCOP2,		disR3000A_COP2[_Funct_] DisFInterfaceN)

TdisR3000AF disR3000A[] = {
	disSPECIAL    , disBCOND     , disJ       , disJAL  , disBEQ , disBNE , disBLEZ , disBGTZ ,
	disADDI       , disADDIU     , disSLTI    , disSLTIU, disANDI, disORI , disXORI , disLUI  ,
	disCOP0       , disNULL      , disCOP2    , disNULL , disNULL, disNULL, disNULL , disNULL ,
	disNULL       , disNULL      , disNULL    , disNULL , disNULL, disNULL, disNULL , disNULL ,
	disLB         , disLH        , disLWL     , disLW   , disLBU , disLHU , disLWR  , disNULL ,
	disSB         , disSH        , disSWL     , disSW   , disNULL, disNULL, disSWR  , disNULL , 
	disNULL       , disNULL      , disNULL    , disNULL , disNULL, disNULL, disNULL , disNULL ,
	disNULL       , disNULL      , disSWC2    , disLWC2 , disNULL, disNULL, disNULL , disNULL };

MakeDisF(disR3000AF,			disR3000A[code >> 26] DisFInterfaceN)
