/*  Pcsx - Pc Psx Emulator
 *  Copyright (C) 1999-2002  Pcsx Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "Common.h"

// Dma0/1 in Mdec.c
// Dma3   in CdRom.c

void hwDma4(unsigned long madr, unsigned long bcr, unsigned long chcr) { // SPU
	switch (chcr) {
		case 0x01000201: //cpu to spu transfer
#ifdef DMA_LOG
			DMA_LOG("*** DMA 4 - SPU mem2spu *** %lx addr = %lx size = %lx\n", chcr, madr, bcr);
#endif
			SPU_writeDMAMem((unsigned short *)&RAM[madr & 0x1fffff], (bcr >> 16) * (bcr & 0xffff) * 2);
			break;
		case 0x01000200: //spu to cpu transfer
#ifdef DMA_LOG
			DMA_LOG("*** DMA 4 - SPU spu2mem *** %lx addr = %lx size = %lx\n", chcr, madr, bcr);
#endif
    		SPU_readDMAMem ((unsigned short *)&RAM[madr & 0x1fffff], (bcr >> 16) * (bcr & 0xffff) * 2);
			break;
#ifdef DMA_LOG
		default:
			DMA_LOG("*** DMA 4 - SPU unknown *** %lx addr = %lx size = %lx\n", chcr, madr, bcr);
			break;
#endif
	}
}

void hwDma2(unsigned long madr, unsigned long bcr, unsigned long chcr) { // GPU
	switch(chcr) {
		case 0x01000200: // vram2mem
#ifdef DMA_LOG
			DMA_LOG("*** DMA 2 - GPU vram2mem *** %lx addr = %lx size = %lx\n", chcr, madr, bcr);
#endif
			GPU_readDataMem((unsigned long *)&RAM[madr & 0x1fffff], (bcr >> 16) * (bcr & 0xffff));
			Cpu->Clear(madr, ((bcr >> 16) * (bcr & 0xffff)) / 4);
			break;

		case 0x01000201: // mem2vram
#ifdef DMA_LOG
			DMA_LOG("*** DMA 2 - GPU mem2vram *** %lx addr = %lx size = %lx\n", chcr, madr, bcr);
#endif
			GPU_writeDataMem((unsigned long *)&RAM[madr & 0x1fffff], (bcr >> 16) * (bcr & 0xffff));
			break;

		case 0x01000401: // dma chain
#ifdef DMA_LOG
			DMA_LOG("*** DMA 2 - GPU dma chain *** %lx addr = %lx size = %lx\n", chcr, madr, bcr);
#endif
			GPU_dmaChain((unsigned long *)RAM, madr & 0x1fffff);
			break;
#ifdef DMA_LOG
		default:
			DMA_LOG("*** DMA 2 - GPU unknown *** %lx addr = %lx size = %lx\n", chcr, madr, bcr);
			break;
#endif
	}
}

void hwDma6(unsigned long madr, unsigned long bcr, unsigned long chcr) {
	unsigned long *mem = (unsigned long *)&RAM[madr & 0x1fffff];

#ifdef DMA_LOG
	DMA_LOG("*** DMA 6 - OT *** %lx addr = %lx size = %lx\n", chcr, madr, bcr);
#endif

	if (chcr == 0x11000002) {
		while (bcr--) {
			*mem-- = (madr - 4) & 0xffffff;
			madr -= 4;
		}
		mem++; *mem = 0xffffff;
	} else {
		// Unknown option
#ifdef DMA_LOG
		DMA_LOG("*** DMA 6 - OT unknown *** %lx addr = %lx size = %lx\n", chcr, madr, bcr);
#endif
	}
}
