/*  Pcsx - Pc Psx Emulator
 *  Copyright (C) 1999-2002  Pcsx Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include "Common.h"
#include "Gte.h"
#include "Memory.h"
#include "Debug.h"
#include "Sio.h"

#ifdef CPU_LOG
//#define DEBUG
#endif

static int branch = 0;
static unsigned long branchPC;

// These macros are used to assemble the repassembler functions

#ifdef DEBUG
#define debugI() \
	if (Log) { \
		CPU_LOG("%s\n", disR3000AF(cpuRegs.code, cpuRegs.pc)); \
	}
#else
#define debugI()
#endif

#define execI() { \
	cpuRegs.code = PC_MEM32(cpuRegs.pc); \
 \
	debugI(); \
 \
	cpuRegs.pc += 4; \
	cpuIrq.cycle++; \
	cpuRegs.GPR.r[0] = 0; \
 \
	intBSC[cpuRegs.code >> 26](); \
}

#define doBranch(tar) { \
	branch = 1; \
	branchPC = tar; \
	execI(); \
	branch = 0; \
	cpuRegs.pc = branchPC; \
 \
	cpuBranchTest(); \
}

// Subsets
void (*intBSC[64])();
void (*intSPC[64])();
void (*intREG[32])();
void (*intCP0[32])();
void (*intCP2[64])();
void (*intCP2BSC[32])();

/*********************************************************
* Arithmetic with immediate operand                      *
* Format:  OP rt, rs, immediate                          *
*********************************************************/
void intADDI() 	{ _rRt_ = _u32(_rRs_) + _Imm_ ; }		// Rt = Rs + Im 	(Exception on Integer Overflow)
void intADDIU() { _rRt_ = _u32(_rRs_) + _Imm_ ; }		// Rt = Rs + Im
void intANDI() 	{ _rRt_ = _u32(_rRs_) & _ImmU_; }		// Rt = Rs And Im
void intORI() 	{ _rRt_ = _u32(_rRs_) | _ImmU_; }		// Rt = Rs Or  Im
void intXORI() 	{ _rRt_ = _u32(_rRs_) ^ _ImmU_; }		// Rt = Rs Xor Im
void intSLTI() 	{ _rRt_ = _i32(_rRs_) < _Imm_ ; }		// Rt = Rs < Im		(Signed)
void intSLTIU() { _rRt_ = _u32(_rRs_) < _ImmU_; }		// Rt = Rs < Im		(Unsigned)

/*********************************************************
* Register arithmetic                                    *
* Format:  OP rd, rs, rt                                 *
*********************************************************/
void intADD()	{ _rRd_ = _u32(_rRs_) + _u32(_rRt_); }	// Rd = Rs + Rt		(Exception on Integer Overflow)
void intADDU() 	{ _rRd_ = _u32(_rRs_) + _u32(_rRt_); }	// Rd = Rs + Rt
void intSUB() 	{ _rRd_ = _u32(_rRs_) - _u32(_rRt_); }	// Rd = Rs - Rt		(Exception on Integer Overflow)
void intSUBU() 	{ _rRd_ = _u32(_rRs_) - _u32(_rRt_); }	// Rd = Rs - Rt
void intAND() 	{ _rRd_ = _u32(_rRs_) & _u32(_rRt_); }	// Rd = Rs And Rt
void intOR() 	{ _rRd_ = _u32(_rRs_) | _u32(_rRt_); }	// Rd = Rs Or  Rt
void intXOR() 	{ _rRd_ = _u32(_rRs_) ^ _u32(_rRt_); }	// Rd = Rs Xor Rt
void intNOR() 	{ _rRd_ =~(_u32(_rRs_) | _u32(_rRt_)); }// Rd = Rs Nor Rt
void intSLT() 	{ _rRd_ = _i32(_rRs_) < _i32(_rRt_); }	// Rd = Rs < Rt		(Signed)
void intSLTU() 	{ _rRd_ = _u32(_rRs_) < _u32(_rRt_); }	// Rd = Rs < Rt		(Unsigned)

/*********************************************************
* Register mult/div & Register trap logic                *
* Format:  OP rs, rt                                     *
*********************************************************/
void intDIV() {
	if (_i32(_rRt_) != 0) {
		_i32(_rLo_) = _i32(_rRs_) / _i32(_rRt_);
		_i32(_rHi_) = _i32(_rRs_) % _i32(_rRt_);
	}
}

void intDIVU() {
	if (_rRt_ != 0) {
		_rLo_ = _rRs_ / _rRt_;
		_rHi_ = _rRs_ % _rRt_;
	}
}

void intMULT() {
	u64 res = (s64)((s64)(_i32(_rRs_)) * (s64)(_i32(_rRt_)));

	cpuRegs.GPR.n.lo = (unsigned long)(res & 0xffffffff);
	cpuRegs.GPR.n.hi = (unsigned long)((res >> 32) & 0xffffffff);
}

void intMULTU() {
	u64 res = (u64)(_u32(_rRs_)) * (u64)(_u32(_rRt_));

	cpuRegs.GPR.n.lo = (unsigned long)(res & 0xffffffff);
	cpuRegs.GPR.n.hi = (unsigned long)((res >> 32) & 0xffffffff);
}

/*********************************************************
* Register branch logic                                  *
* Format:  OP rs, offset                                 *
*********************************************************/
#define RepZBranchi32(op)      if(_i32(_rRs_) op 0) doBranch(_BranchTarget_);
#define RepZBranchLinki32(op)  if(_i32(_rRs_) op 0) { _SetLink(31); doBranch(_BranchTarget_); }

void intBGEZ()   { RepZBranchi32(>=) }      // Branch if Rs >= 0
void intBGEZAL() { RepZBranchLinki32(>=) }  // Branch if Rs >= 0 and link
void intBGTZ()   { RepZBranchi32(>) }       // Branch if Rs >  0
void intBLEZ()   { RepZBranchi32(<=) }      // Branch if Rs <= 0
void intBLTZ()   { RepZBranchi32(<) }       // Branch if Rs <  0
void intBLTZAL() { RepZBranchLinki32(<) }   // Branch if Rs <  0 and link

/*********************************************************
* Shift arithmetic with constant shift                   *
* Format:  OP rd, rt, sa                                 *
*********************************************************/
void intSLL() { _u32(_rRd_) = _u32(_rRt_) << _Sa_; } // Rd = Rt << sa
void intSRA() { _i32(_rRd_) = _i32(_rRt_) >> _Sa_; } // Rd = Rt >> sa (arithmetic)
void intSRL() { _u32(_rRd_) = _u32(_rRt_) >> _Sa_; } // Rd = Rt >> sa (logical)

/*********************************************************
* Shift arithmetic with variant register shift           *
* Format:  OP rd, rt, rs                                 *
*********************************************************/
void intSLLV() { _u32(_rRd_) = _u32(_rRt_) << _u32(_rRs_); } // Rd = Rt << rs
void intSRAV() { _i32(_rRd_) = _i32(_rRt_) >> _u32(_rRs_); } // Rd = Rt >> rs (arithmetic)
void intSRLV() { _u32(_rRd_) = _u32(_rRt_) >> _u32(_rRs_); } // Rd = Rt >> rs (logical)

/*********************************************************
* Load higher 16 bits of the first word in GPR with imm  *
* Format:  OP rt, immediate                              *
*********************************************************/
void intLUI() { _u32(_rRt_) = cpuRegs.code << 16; } // Upper halfword of Rt = Im

/*********************************************************
* Move from HI/LO to GPR                                 *
* Format:  OP rd                                         *
*********************************************************/
void intMFHI() { _rRd_ = _rHi_; } // Rd = Hi
void intMFLO() { _rRd_ = _rLo_; } // Rd = Lo

/*********************************************************
* Move to GPR to HI/LO & Register jump                   *
* Format:  OP rs                                         *
*********************************************************/
void intMTHI() { _rHi_ = _rRs_; } // Hi = Rs
void intMTLO() { _rLo_ = _rRs_; } // Lo = Rs

/*********************************************************
* Special purpose instructions                           *
* Format:  OP                                            *
*********************************************************/
void intBREAK() { // Break exception
	cpuRegs.pc -= 4;
	cpuTakeException(0x40, branch);
}

void intSYSCALL() {
	cpuRegs.pc -= 4;
	cpuTakeException(0x20, branch);
}

void intRFE() {
	cpuRegs.CP0.n.Status = (cpuRegs.CP0.n.Status & 0xfffffff0) |
						  ((cpuRegs.CP0.n.Status & 0x3c) >> 2);
}

/*********************************************************
* Register branch logic                                  *
* Format:  OP rs, rt, offset                             *
*********************************************************/
#define RepBranchi32(op)      if(_i32(_rRs_) op _i32(_rRt_)) doBranch(_BranchTarget_);

void intBEQ() {	RepBranchi32(==) }  // Branch if Rs == Rt
void intBNE() {	RepBranchi32(!=) }  // Branch if Rs != Rt

/*********************************************************
* Jump to target                                         *
* Format:  OP target                                     *
*********************************************************/
void intJ()   {               doBranch(_JumpTarget_); }
void intJAL() {	_SetLink(31); doBranch(_JumpTarget_); }

/*********************************************************
* Register jump                                          *
* Format:  OP rs, rd                                     *
*********************************************************/
void intJR()   {                 doBranch(_u32(_rRs_)); }
void intJALR() { _SetLink(_Rd_); doBranch(_u32(_rRs_)); }

/*********************************************************
* Load and store for GPR                                 *
* Format:  OP rt, offset(base)                           *
*********************************************************/

#define _oB_ (_u32(_rRs_) + _Imm_)

void intLB()  { _i32(_rRt_) = (char)memRead8(_oB_); }
void intLBU() { _u32(_rRt_) = memRead8(_oB_); }
void intLH()  { _i32(_rRt_) = (short)memRead16(_oB_); }
void intLHU() { _u32(_rRt_) = memRead16(_oB_); }
void intLW()  { _u32(_rRt_) = memRead32(_oB_); }

void intLWL() {
	u32 shift = (_oB_ & 3) << 3;
	u32 mem = memRead32(_oB_ & 0xfffffffc);

	_u32(_rRt_) =	( _u32(_rRt_) & (0x00ffffff >> shift) ) | 
					( mem << (24 - shift) );

	/*
	Mem = 1234.  Reg = abcd

	0   4bcd   (mem << 24) | (reg & 0x00ffffff)
	1   34cd   (mem << 16) | (reg & 0x0000ffff)
	2   234d   (mem <<  8) | (reg & 0x000000ff)
	3   1234   (mem      ) | (reg & 0x00000000)

	*/
}

void intLWR() {
	u32 shift = (_oB_ & 3) << 3;
	u32 mem = memRead32(_oB_ & 0xfffffffc);

	_u32(_rRt_) =	( _u32(_rRt_) & (0xffffff00 << (24 - shift)) ) |
					( mem  >> shift );

	/*
	Mem = 1234.  Reg = abcd

	0   1234   (mem      ) | (reg & 0x00000000)
	1   a123   (mem >>  8) | (reg & 0xff000000)
	2   ab12   (mem >> 16) | (reg & 0xffff0000)
	3   abc1   (mem >> 24) | (reg & 0xffffff00)

	*/
}

void intSB() { memWrite8 (_oB_, _u8 (_rRt_)); }
void intSH() { memWrite16(_oB_, _u16(_rRt_)); }
void intSW() { memWrite32(_oB_, _u32(_rRt_)); }


void intSWL() {
	u32 shift = (_oB_ & 3) << 3;
	u32 mem = memRead32(_oB_ & 0xfffffffc);

	memWrite32((_oB_ & 0xfffffffc),  ( ( _u32(_rRt_) >>  (24 - shift) ) ) |
			  (  mem & (0xffffff00 << shift) ));
	/*
	Mem = 1234.  Reg = abcd

	0   123a   (reg >> 24) | (mem & 0xffffff00)
	1   12ab   (reg >> 16) | (mem & 0xffff0000)
	2   1abc   (reg >>  8) | (mem & 0xff000000)
	3   abcd   (reg      ) | (mem & 0x00000000)

	*/
}

void intSWR() {
	u32 shift = (_oB_ & 3) << 3;
	u32 mem = memRead32(_oB_ & 0xfffffffc);

	memWrite32((_oB_ & 0xfffffffc), ( ( _u32(_rRt_) << shift ) |
			   (mem  & (0x00ffffff >> (24 - shift)) ) ) );
	/*
	Mem = 1234.  Reg = abcd

	0   abcd   (reg      ) | (mem & 0x00000000)
	1   bcd4   (reg <<  8) | (mem & 0x000000ff)
	2   cd34   (reg << 16) | (mem & 0x0000ffff)
	3   d234   (reg << 24) | (mem & 0x00ffffff)

	*/
}


/*********************************************************
* Load and store for COP2                                *
* Format:  OP rt, offset(base)                           *
*********************************************************/

//void intLWC2() { cpuRegs.CP2D.r[_Rt_] = memRead32(_oB_); }
//void intSWC2() { memWrite32(_oB_, cpuRegs.CP2D.r[_Rt_]); }

/*********************************************************
* Moves between GPR and COPx                             *
* Format:  OP rt, fs                                     *
*********************************************************/
void intMFC0() { _i32(_rRt_) = (int)_rFs_; }
void intCFC0() { _i32(_rRt_) = (int)_rFs_; }

void intMTC0() { _rFs_ = _u32(_rRt_); }
void intCTC0() { _rFs_ = _u32(_rRt_); }

/*********************************************************
* Unknow instruction (would generate an exception)       *
* Format:  ?                                             *
*********************************************************/
void intNULL() { 
#ifdef CPU_LOG
	CPU_LOG("int: Unimplemented op %x\n", cpuRegs.code);
#endif
}

void intSPECIAL() {
	intSPC[_Funct_]();
}

void intREGIMM() {
	intREG[_Rt_]();
}

void intCOP0() {
	intCP0[_Rs_]();
}

void intCOP2() {
	intCP2[_Funct_]();
}

void intBASIC() {
	intCP2BSC[_Rs_]();
}

void (*intBSC[64])() = {
	intSPECIAL, intREGIMM, intJ   , intJAL  , intBEQ , intBNE , intBLEZ, intBGTZ,
	intADDI   , intADDIU , intSLTI, intSLTIU, intANDI, intORI , intXORI, intLUI ,
	intCOP0   , intNULL  , intCOP2, intNULL , intNULL, intNULL, intNULL, intNULL,
	intNULL   , intNULL  , intNULL, intNULL , intNULL, intNULL, intNULL, intNULL,
	intLB     , intLH    , intLWL , intLW   , intLBU , intLHU , intLWR , intNULL,
	intSB     , intSH    , intSWL , intSW   , intNULL, intNULL, intSWR , intNULL, 
	intNULL   , intNULL  , cp2LWC2, intNULL , intNULL, intNULL, intNULL, intNULL,
	intNULL   , intNULL  , cp2SWC2, intNULL , intNULL, intNULL, intNULL, intNULL 
};


void (*intSPC[64])() = {
	intSLL , intNULL , intSRL , intSRA , intSLLV   , intNULL , intSRLV, intSRAV,
	intJR  , intJALR , intNULL, intNULL, intSYSCALL, intBREAK, intNULL, intNULL,
	intMFHI, intMTHI , intMFLO, intMTLO, intNULL   , intNULL , intNULL, intNULL,
	intMULT, intMULTU, intDIV , intDIVU, intNULL   , intNULL , intNULL, intNULL,
	intADD , intADDU , intSUB , intSUBU, intAND    , intOR   , intXOR , intNOR ,
	intNULL, intNULL , intSLT , intSLTU, intNULL   , intNULL , intNULL, intNULL,
	intNULL, intNULL , intNULL, intNULL, intNULL   , intNULL , intNULL, intNULL,
	intNULL, intNULL , intNULL, intNULL, intNULL   , intNULL , intNULL, intNULL
};

void (*intREG[32])() = {
	intBLTZ  , intBGEZ  , intNULL, intNULL, intNULL, intNULL, intNULL, intNULL,
	intNULL  , intNULL  , intNULL, intNULL, intNULL, intNULL, intNULL, intNULL,
	intBLTZAL, intBGEZAL, intNULL, intNULL, intNULL, intNULL, intNULL, intNULL,
	intNULL  , intNULL  , intNULL, intNULL, intNULL, intNULL, intNULL, intNULL
};

void (*intCP0[32])() = {
	intMFC0, intNULL, intCFC0, intNULL, intMTC0, intNULL, intCTC0, intNULL,
	intNULL, intNULL, intNULL, intNULL, intNULL, intNULL, intNULL, intNULL,
	intRFE , intNULL, intNULL, intNULL, intNULL, intNULL, intNULL, intNULL,
	intNULL, intNULL, intNULL, intNULL, intNULL, intNULL, intNULL, intNULL
};

void (*intCP2[64])() = {
	intBASIC, cp2RTPS , intNULL , intNULL, intNULL, intNULL , cp2NCLIP, intNULL, // 00
	intNULL , intNULL , intNULL , intNULL, cp2OP  , intNULL , intNULL , intNULL, // 08
	cp2DPCS , cp2INTPL, cp2MVMVA, cp2NCDS, cp2CDP , intNULL , cp2NCDT , intNULL, // 10
	intNULL , intNULL , intNULL , cp2NCCS, cp2CC  , intNULL , cp2NCS  , intNULL, // 18
	cp2NCT  , intNULL , intNULL , intNULL, intNULL, intNULL , intNULL , intNULL, // 20
	cp2SQR  , cp2DCPL , cp2DPCT , intNULL, intNULL, cp2AVSZ3, cp2AVSZ4, intNULL, // 28 
	cp2RTPT , intNULL , intNULL , intNULL, intNULL, intNULL , intNULL , intNULL, // 30
	intNULL , intNULL , intNULL , intNULL, intNULL, cp2GPF  , cp2GPL  , cp2NCCT  // 38
};

void (*intCP2BSC[32])() = {
	cp2MFC2, intNULL, cp2CFC2, intNULL, cp2MTC2, intNULL, cp2CTC2, intNULL,
	intNULL, intNULL, intNULL, intNULL, intNULL, intNULL, intNULL, intNULL,
	intNULL, intNULL, intNULL, intNULL, intNULL, intNULL, intNULL, intNULL,
	intNULL, intNULL, intNULL, intNULL, intNULL, intNULL, intNULL, intNULL
};


///////////////////////////////////////////

int intInit() {
	return 0;
}

void intReset() {
}

void intExecute() {
	for (;;) execI();
}

void intExecuteBios() {
	while (cpuRegs.pc != 0x80030000)
		execI();
}

void intClear(u32 Addr, u32 Size) {
}

void intShutdown() {
}

R3000Acpu intCpu = {
	intInit,
	intReset,
	intExecuteBios,
	intExecute,
	intClear,
	intShutdown
};
