/*  Pcsx - Pc Psx Emulator
 *  Copyright (C) 1999-2002  Pcsx Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <string.h>

#include "Common.h"
#include "Memory.h"
#include "device.h"
#include "Debug.h"
#include "Sio.h"

int cpuInit() {

	if (Config.Cpu) Cpu = &intCpu;
#ifdef __i386__
	else Cpu = &recCpu;
#endif
//	Log=1;

	if (memInit() == -1) return -1;

	return Cpu->Init();
}

void cpuReset() {

	Cpu->Reset();

	memReset();

	memcpy(HWM, self_sram, 0x1000);
	memset(&cpuRegs, 0, sizeof(cpuRegs));
	memset(&cpuIrq , 0, sizeof(cpuIrq));

	cpuRegs.pc = 0xbfc00000; // Start in bootstrap

	cpuRegs.CP0.r[12] = 0x10900000; // COP0 enabled | BEV = 1 | TS = 1
	cpuRegs.CP0.r[15] = 0x00000002; // PRevID = Revision ID, same as R3000A

	hwReset();
	biosInit();

	if (!Config.HLE) Cpu->ExecuteBios();

#ifdef CPU_LOG
	CPU_LOG("*BIOS END*\n");
#endif
	Log=0;
}

void cpuShutdown() {
	memShutdown();
	biosShutdown();

	Cpu->Shutdown();
}

void cpuTakeException(unsigned long code, unsigned long bd) {
	// Set the Cause
	cpuRegs.CP0.n.Cause = code;

#ifdef CPU_LOG
	if (bd) CPU_LOG("bd set\n");
#endif
	// Set the EPC & PC
	if (bd) {
		cpuRegs.CP0.n.Cause|= 0x80000000;
		cpuRegs.CP0.n.EPC = (cpuRegs.pc - 4);
	} else
		cpuRegs.CP0.n.EPC = (cpuRegs.pc);

	if (cpuRegs.CP0.n.Status & 0x400000)
		cpuRegs.pc = 0xbfc00180;
	else
		cpuRegs.pc = 0x80000080;

	// Set the Status
	cpuRegs.CP0.n.Status = (cpuRegs.CP0.n.Status &~0x3f) |
						  ((cpuRegs.CP0.n.Status & 0xf) << 2);

	if (Config.HLE) biosException();
}

void cpuBranchTest() {

	if ((cpuIrq.cycle - nextsCounter) >= nextCounter)
		rcntUpdate();

	if (cpuIrq.interrupt) {
		if ((cpuIrq.interrupt & 0x80) && (!Config.Sio)) { // sio
			if ((cpuIrq.cycle - cpuIrq.intCycle[7]) >= cpuIrq.intCycle[7+1]) {
				cpuIrq.interrupt&=~0x80;
				sioInterrupt();
			}
		}
		if (cpuIrq.interrupt & 0x04) { // cdr
			if ((cpuIrq.cycle - cpuIrq.intCycle[2]) >= cpuIrq.intCycle[2+1]) {
				cpuIrq.interrupt&=~0x04;
				cdrInterrupt();
			}
		}
		if (cpuIrq.interrupt & 0x040000) { // cdr read
			if ((cpuIrq.cycle - cpuIrq.intCycle[2+16]) >= cpuIrq.intCycle[2+16+1]) {
				cpuIrq.interrupt&=~0x040000;
				cdrReadInterrupt();
			}
		}
	}

	if (uHWM32(0x1070) & uHWM32(0x1074)) {
		if ((cpuRegs.CP0.n.Status & 0x401) == 0x401) {
#ifdef CPU_LOG
//			CPU_LOG("Interrupt: %x  %x\n", uHWM32(0x1070), uHWM32(0x1074));
#endif
			cpuTakeException(0x400, 0);
		}
	}

	if (Config.PsxOut || Config.HLE) {
		switch (cpuRegs.pc & 0x1fffff) {
			case 0x80:
				if (Config.HLE) biosException();
				break;
		    case 0xa0:
#ifdef BIOS_LOG
				BIOS_LOG("Bios call a0: %s (%x)\n", biosA0n[cpuRegs.GPR.n.t1 & 0xff], cpuRegs.GPR.n.t1 & 0xff);
#endif
				if (biosA0[cpuRegs.GPR.n.t1 & 0xff])
			   		biosA0[cpuRegs.GPR.n.t1 & 0xff]();
			   	break;
		    case 0xb0:
#ifdef BIOS_LOG
				BIOS_LOG("Bios call b0: %s (%x)\n", biosB0n[cpuRegs.GPR.n.t1 & 0xff], cpuRegs.GPR.n.t1 & 0xff);
#endif
				if (biosB0[cpuRegs.GPR.n.t1 & 0xff])
			   		biosB0[cpuRegs.GPR.n.t1 & 0xff]();
			   	break;
			case 0xc0:
#ifdef BIOS_LOG
				BIOS_LOG("Bios call c0: %s (%x)\n", biosC0n[cpuRegs.GPR.n.t1 & 0xff], cpuRegs.GPR.n.t1 & 0xff);
#endif
				if (biosC0[cpuRegs.GPR.n.t1 & 0xff])
		    		biosC0[cpuRegs.GPR.n.t1 & 0xff]();
		    	break;
		}
	}
}

